import json
import traceback
import urwid
import webbrowser

from toot import __version__
from toot import api

from toot.tui.utils import highlight_keys
from toot.tui.images import image_support_enabled, load_image, graphics_widget
from toot.tui.widgets import Button, EditBox, SelectableText
from toot.tui.richtext import html_to_widgets


class StatusSource(urwid.Padding):
    """Shows status data, as returned by the server, as formatted JSON."""
    def __init__(self, status):
        self.source = json.dumps(status.data, indent=4)
        self.filename_edit = EditBox(caption="Filename: ", edit_text=f"status-{status.id}.json")
        self.status_text = urwid.Text("")

        walker = urwid.SimpleFocusListWalker([
            self.filename_edit,
            Button("Save", on_press=self.save_json),
            urwid.Divider("─"),
            urwid.Divider(" "),
            urwid.Text(self.source)
        ])

        frame = urwid.Frame(
            body=urwid.ListBox(walker),
            footer=self.status_text
        )
        super().__init__(frame)

    def save_json(self, button):
        filename = self.filename_edit.get_edit_text()
        if filename:
            with open(filename, "w") as f:
                f.write(self.source)
            self.status_text.set_text(("footer_message", f"Saved to {filename}"))


class StatusZoom(urwid.ListBox):
    """Opens status in scrollable popup window"""
    def __init__(self, status_details):
        ll = list(filter(lambda x: getattr(x, "rows", None), status_details.widget_list))
        walker = urwid.SimpleFocusListWalker(ll)
        super().__init__(walker)


class StatusLinks(urwid.ListBox):
    """Shows status links."""
    signals = ["clear-screen"]

    def __init__(self, links):

        def widget(url, title):
            return Button(title or url, on_press=lambda btn: self.browse(url))

        walker = urwid.SimpleFocusListWalker(
            [widget(url, title) for url, title in links]
        )
        super().__init__(walker)

    def browse(self, url):
        webbrowser.open(url)
        # force a screen refresh; necessary with console browsers
        self._emit("clear-screen")


class ExceptionStackTrace(urwid.ListBox):
    """Shows an exception stack trace."""
    def __init__(self, ex):
        lines = traceback.format_exception(type(ex), value=ex, tb=ex.__traceback__)
        walker = urwid.SimpleFocusListWalker([
            urwid.Text(line) for line in lines
        ])
        super().__init__(walker)


class StatusDeleteConfirmation(urwid.ListBox):
    signals = ["delete", "close"]

    def __init__(self, status):
        def _delete(_):
            self._emit("delete")

        def _close(_):
            self._emit("close")

        walker = urwid.SimpleFocusListWalker([
            Button("Yes, delete", on_press=_delete),
            Button("No, cancel", on_press=_close),
        ])
        super().__init__(walker)


class GotoMenu(urwid.ListBox):
    signals = [
        "home_timeline",
        "public_timeline",
        "hashtag_timeline",
        "bookmark_timeline",
        "notification_timeline",
        "conversation_timeline",
        "personal_timeline",
        "list_timeline",
    ]

    def __init__(self, user_timelines, user_lists):
        self.hash_edit = EditBox(caption="Hashtag: ")
        self.message_widget = urwid.Text("")

        actions = list(self.generate_actions(user_timelines, user_lists))
        walker = urwid.SimpleFocusListWalker(actions)
        super().__init__(walker)

    def get_hashtag(self):
        return self.hash_edit.edit_text.strip().lstrip("#")

    def generate_actions(self, user_timelines, user_lists):
        def _home(button):
            self._emit("home_timeline")

        def _local_public(button):
            self._emit("public_timeline", True)

        def _global_public(button):
            self._emit("public_timeline", False)

        def _personal(button):
            self._emit("personal_timeline", False)

        def _bookmarks(button):
            self._emit("bookmark_timeline", False)

        def _notifications(button):
            self._emit("notification_timeline", False)

        def _conversations(button):
            self._emit("conversation_timeline", False)

        def _hashtag(local):
            self.message_widget.set_text("")
            hashtag = self.get_hashtag()
            if hashtag:
                self._emit("hashtag_timeline", hashtag, local)
            else:
                self.message_widget.set_text(("warning", "Hashtag name required"))

        def mk_on_press_user_hashtag(tag, local):
            def on_press(btn):
                self._emit("hashtag_timeline", tag, local)
            return on_press

        def mk_on_press_user_list(list_item):
            def on_press(btn):
                self._emit("list_timeline", list_item)
            return on_press

        yield Button("Home timeline", on_press=_home)
        yield Button("Local public timeline", on_press=_local_public)
        yield Button("Global public timeline", on_press=_global_public)
        yield Button("Personal timeline", on_press=_personal)
        yield Button("Bookmarks", on_press=_bookmarks)
        yield Button("Notifications", on_press=_notifications)
        yield Button("Conversations", on_press=_conversations)

        if len(user_timelines):
            yield urwid.Divider()
            yield urwid.Text(("bold", "Shortcuts:"))

            # show all hashtag shortcuts
            for tag, cfg in sorted(user_timelines.items()):
                is_local = cfg["local"]
                yield Button(f"#{tag}" + (" (local)" if is_local else ""),
                     on_press=mk_on_press_user_hashtag(tag, is_local))

        for list_item in user_lists:
            yield Button(f"\N{clipboard}{list_item['title']}",
                         on_press=mk_on_press_user_list(list_item))

        yield urwid.Divider()
        yield self.hash_edit
        yield Button("Local hashtag timeline", on_press=lambda x: _hashtag(True))
        yield Button("Public hashtag timeline", on_press=lambda x: _hashtag(False))
        yield urwid.Divider()
        yield self.message_widget


class Help(urwid.Padding):
    def __init__(self):
        actions = list(self.generate_contents())
        walker = urwid.SimpleListWalker(actions)
        listbox = urwid.ListBox(walker)
        super().__init__(listbox, left=1, right=1)

    def generate_contents(self):
        def h(text):
            return highlight_keys(text, "shortcut")

        yield urwid.Text(("bold", "toot {}".format(__version__)))
        yield urwid.Divider()
        yield urwid.Text(("bold", "General usage"))
        yield urwid.Divider()
        yield urwid.Text(h("  [Arrow keys] or [h/j/k/l] to move around and scroll content"))
        yield urwid.Text(h("  [PageUp] and [PageDown] to scroll content"))
        yield urwid.Text(h("  [Enter] or [Space] to activate buttons and menu options"))
        yield urwid.Text(h("  [Esc] or [q] to go back, close overlays, such as menus and this help text"))
        yield urwid.Divider()
        yield urwid.Text(("bold", "General keys"))
        yield urwid.Divider()
        yield urwid.Text(h("  [q] - quit toot"))
        yield urwid.Text(h("  [g] - go to - switch timelines"))
        yield urwid.Text(h("  [P] - pin/unpin current timeline"))
        yield urwid.Text(h("  [,] - refresh current timeline"))
        yield urwid.Text(h("  [?] - show this help"))
        yield urwid.Divider()
        yield urwid.Text(("bold", "Status keys"))
        yield urwid.Divider()
        yield urwid.Text("These commands are applied to the currently focused status.")
        yield urwid.Divider()
        yield urwid.Text(h("  [a] - Show account"))
        yield urwid.Text(h("  [b] - Boost/unboost status"))
        yield urwid.Text(h("  [c] - Compose a new status"))
        yield urwid.Text(h("  [d] - Delete status"))
        yield urwid.Text(h("  [e] - Edit status"))
        yield urwid.Text(h("  [f] - Favourite/unfavourite status"))
        yield urwid.Text(h("  [i] - Show the status links"))
        yield urwid.Text(h("  [m] - Show media in image viewer"))
        yield urwid.Text(h("  [n] - Translate status if possible (toggle)"))
        yield urwid.Text(h("  [o] - Bookmark/unbookmark status"))
        yield urwid.Text(h("  [p] - Vote on a poll"))
        yield urwid.Text(h("  [r] - Reply to status"))
        yield urwid.Text(h("  [s] - Show text marked as sensitive"))
        yield urwid.Text(h("  [t] - Show status thread (replies)"))
        yield urwid.Text(h("  [u] - Show the status data in JSON as received from the server"))
        yield urwid.Text(h("  [v] - Open status in default browser"))
        yield urwid.Text(h("  [y] - Copy status to clipboard"))
        yield urwid.Text(h("  [z] - Open status in scrollable popup window"))
        yield urwid.Divider()
        yield urwid.Text(("bold", "Links"))
        yield urwid.Divider()
        yield link("Documentation: ", "https://toot.bezdomni.net/")
        yield link("Project home:  ", "https://github.com/ihabunek/toot/")


class Account(urwid.ListBox):
    """Shows account data and provides various actions"""
    def __init__(self, app, user, account, relationship, options):
        self.app = app
        self.user = user
        self.account = account
        self.relationship = relationship
        self.options = options
        self.last_action = None
        self.setup_listbox()

    def setup_listbox(self):
        actions = list(self.generate_contents(self.account, self.relationship, self.last_action))
        walker = urwid.SimpleListWalker(actions)
        super().__init__(walker)

    def account_header(self, account):
        if image_support_enabled() and account['avatar'] and not account["avatar"].endswith("missing.png"):
            img = load_image(account['avatar'])
            aimg = urwid.BoxAdapter(
                graphics_widget(img, image_format=self.options.image_format, corner_radius=10,
                                colors=self.options.colors), 10)
        else:
            aimg = urwid.BoxAdapter(urwid.SolidFill(" "), 10)

        if image_support_enabled() and account['header'] and not account["header"].endswith("missing.png"):
            img = load_image(account['header'])

            himg = (urwid.BoxAdapter(
                graphics_widget(img, image_format=self.options.image_format, corner_radius=10,
                                colors=self.options.colors), 10))
        else:
            himg = urwid.BoxAdapter(urwid.SolidFill(" "), 10)

        atxt = urwid.Pile([urwid.Divider(),
                           (urwid.Text(("account", account["display_name"]))),
                (urwid.Text(("highlight", "@" + self.account['acct'])))])
        columns = urwid.Columns([aimg, ("weight", 9999, himg)], dividechars=2, min_width=20)

        header = urwid.Pile([columns, urwid.Divider(), atxt])
        return header

    def generate_contents(self, account, relationship=None, last_action=None):
        if self.last_action and not self.last_action.startswith("Confirm"):
            yield Button(f"Confirm {self.last_action}", on_press=take_action, user_data=self)
            yield Button("Cancel", on_press=cancel_action, user_data=self)
        else:
            if self.user.username == account["acct"]:
                yield urwid.Text(("dim", "This is your account"))
            else:
                if relationship['requested']:
                    yield urwid.Text(("dim", "< Follow request is pending >"))
                else:
                    yield Button("Unfollow" if relationship['following'] else "Follow",
                    on_press=confirm_action, user_data=self)

                yield Button("Unmute" if relationship['muting'] else "Mute",
                    on_press=confirm_action, user_data=self)
                yield Button("Unblock" if relationship['blocking'] else "Block",
                    on_press=confirm_action, user_data=self)

        yield urwid.Divider("─")
        yield urwid.Divider()

        yield self.account_header(account)

        if account["note"]:
            yield urwid.Divider()
            widgetlist = html_to_widgets(account["note"])
            for line in widgetlist:
                yield (line)

        yield urwid.Divider()
        yield urwid.Text(["ID: ", ("highlight", f"{account['id']}")])
        yield urwid.Text(["Since: ", ("highlight", f"{account['created_at'][:10]}")])
        yield urwid.Divider()

        if account["bot"]:
            yield urwid.Text([("highlight", "Bot \N{robot face}")])
            yield urwid.Divider()
        if account["locked"]:
            yield urwid.Text([("warning", "Locked \N{lock}")])
            yield urwid.Divider()
        if "suspended" in account and account["suspended"]:
            yield urwid.Text([("warning", "Suspended \N{cross mark}")])
            yield urwid.Divider()
        if relationship["followed_by"]:
            yield urwid.Text(("highlight", "Follows you \N{busts in silhouette}"))
            yield urwid.Divider()
        if relationship["blocked_by"]:
            yield urwid.Text(("warning", "Blocks you \N{no entry}"))
            yield urwid.Divider()

        yield urwid.Text(["Followers: ", ("highlight", f"{account['followers_count']}")])
        yield urwid.Text(["Following: ", ("highlight", f"{account['following_count']}")])
        yield urwid.Text(["Statuses: ", ("highlight", f"{account['statuses_count']}")])

        if account["fields"]:
            for field in account["fields"]:
                name = field["name"].title()
                yield urwid.Divider()
                yield urwid.Text([("bold", f"{name.rstrip(':')}"), ":"])

                widgetlist = html_to_widgets(field["value"])
                for line in widgetlist:
                    yield (line)

                if 'verified_at' in field and field["verified_at"]:
                    yield urwid.Text(("success", "✓ Verified"))

        yield urwid.Divider()
        yield link("", account["url"])


def take_action(button: Button, self: Account):
    action = button.get_label()

    if action == "Confirm Follow":
        self.relationship = api.follow(self.app, self.user, self.account["id"]).json()
    elif action == "Confirm Unfollow":
        self.relationship = api.unfollow(self.app, self.user, self.account["id"]).json()
    elif action == "Confirm Mute":
        self.relationship = api.mute(self.app, self.user, self.account["id"]).json()
    elif action == "Confirm Unmute":
        self.relationship = api.unmute(self.app, self.user, self.account["id"]).json()
    elif action == "Confirm Block":
        self.relationship = api.block(self.app, self.user, self.account["id"]).json()
    elif action == "Confirm Unblock":
        self.relationship = api.unblock(self.app, self.user, self.account["id"]).json()

    self.last_action = None
    self.setup_listbox()


def confirm_action(button: Button, self: Account):
    self.last_action = button.get_label()
    self.setup_listbox()


def cancel_action(button: Button, self: Account):
    self.last_action = None
    self.setup_listbox()


def link(text, url):
    attr_map = {"link": "link_focused"}
    text = SelectableText([text, ("link", url)])
    urwid.connect_signal(text, "click", lambda t: webbrowser.open(url))
    return urwid.AttrMap(text, "", attr_map)



import click
import os
import sys

from datetime import datetime, timedelta, timezone
from time import sleep, time
from typing import BinaryIO, Optional, Tuple

from toot import api, config
from toot.cache import get_last_post_id, save_last_post_id
from toot.cli import AccountParamType, cli, json_option, pass_context, Context
from toot.cli import DURATION_EXAMPLES, VISIBILITY_CHOICES
from toot.tui.constants import VISIBILITY_OPTIONS  # move to top-level ?

from toot.cli.validators import validate_duration, validate_language
from toot.entities import MediaAttachment, from_dict
from toot.utils import EOF_KEY, delete_tmp_status_file, editor_input, multiline_input
from toot.utils.datetime import parse_datetime


@cli.command()
@click.argument("text", required=False)
@click.option(
    "--media", "-m",
    help="""Path to media file to attach, can be used multiple times to attach
            multiple files.""",
    type=click.File(mode="rb"),
    multiple=True
)
@click.option(
    "--description", "-d", "descriptions",
    help="""Plain-text description of the media for accessibility purposes, one
            per attached media""",
    multiple=True,
)
@click.option(
    "--thumbnail", "thumbnails",
    help="Path to an image file to serve as media thumbnail, one per attached media",
    type=click.File(mode="rb"),
    multiple=True
)
@click.option(
    "--visibility", "-v",
    help="Post visibility: " + "; "
      .join("{} = {}".format(visibility, description)
      for visibility, caption, description in VISIBILITY_OPTIONS),
    type=click.Choice(VISIBILITY_CHOICES),
)
@click.option(
    "--sensitive", "-s",
    help="Mark status and attached media as sensitive",
    default=False,
    is_flag=True,
)
@click.option(
    "--spoiler-text", "-p",
    help="Text to be shown as a warning or subject before the actual content.",
)
@click.option(
    "--reply-to", "-r",
    help="ID of the status being replied to, if status is a reply.",
)
@click.option(
    "--reply-last", "-R",
    help="Reply to the last posted status to continue the thread.",
    is_flag=True,
    default=False,
)
@click.option(
    "--language", "-l",
    help="ISO 639-1 language code of the toot, to skip automatic detection.",
    callback=validate_language,
)
@click.option(
    "--editor", "-e",
    is_flag=False,
    flag_value=os.getenv("EDITOR"),
    help="""Specify an editor to compose your toot. When used without a value
            it will use the editor defined in the $EDITOR environment variable.""",
)
@click.option(
    "--scheduled-at",
    help="""ISO 8601 Datetime at which to schedule a status. Must be at least 5
            minutes in the future.""",
)
@click.option(
    "--scheduled-in",
    help=f"""Schedule the toot to be posted after a given amount of time,
             {DURATION_EXAMPLES}. Must be at least 5 minutes.""",
    callback=validate_duration,
)
@click.option(
    "--content-type", "-t",
    help="MIME type for the status text (not supported on all instances)",
)
@click.option(
    "--poll-option",
    help="Possible answer to the poll, can be given multiple times.",
    multiple=True,
)
@click.option(
    "--poll-expires-in",
    help=f"Duration that the poll should be open, {DURATION_EXAMPLES}",
    callback=validate_duration,
    default="24h",
)
@click.option(
    "--poll-multiple",
    help="Allow multiple answers to be selected.",
    is_flag=True,
    default=False,
)
@click.option(
    "--poll-hide-totals",
    help="Hide vote counts until the poll ends.",
    is_flag=True,
    default=False,
)
@click.option(
    "-u", "--using",
    type=AccountParamType(),
    help="The account to use, overrides the active account.",
)
@json_option
@pass_context
def post(
    ctx: Context,
    text: Optional[str],
    media: Tuple[str],
    descriptions: Tuple[str],
    thumbnails: Tuple[str],
    visibility: Optional[str],
    sensitive: bool,
    spoiler_text: Optional[str],
    reply_to: Optional[str],
    language: Optional[str],
    editor: Optional[str],
    scheduled_at: Optional[str],
    scheduled_in: Optional[int],
    content_type: Optional[str],
    poll_option: Tuple[str],
    poll_expires_in: int,
    poll_multiple: bool,
    poll_hide_totals: bool,
    json: bool,
    using: str,
    reply_last: bool,
):
    """Post a new status"""
    if len(media) > 4:
        raise click.ClickException("Cannot attach more than 4 files.")

    if using:
        user, app = config.get_user_app(using)
        if not user or not app:
            raise click.ClickException(f"Account '{using}' not found. Run `toot auth` to see available accounts.")
    else:
        user, app = ctx.user, ctx.app

    media_ids = _upload_media(app, user, media, descriptions, thumbnails)
    status_text = _get_status_text(text, editor, media)
    scheduled_at = _get_scheduled_at(scheduled_at, scheduled_in)
    reply_to = _get_reply_to(app, user, reply_to, reply_last)

    if not status_text and not media_ids:
        raise click.ClickException("You must specify either text or media to post.")

    response = api.post_status(
        app,
        user,
        status_text,
        visibility=visibility,
        media_ids=media_ids,
        sensitive=sensitive,
        spoiler_text=spoiler_text,
        in_reply_to_id=reply_to,
        language=language,
        scheduled_at=scheduled_at,
        content_type=content_type,
        poll_options=poll_option,
        poll_expires_in=poll_expires_in,
        poll_multiple=poll_multiple,
        poll_hide_totals=poll_hide_totals,
    )

    status = response.json()

    if json:
        click.echo(response.text)
    else:
        if "scheduled_at" in status:
            scheduled_at = parse_datetime(status["scheduled_at"])
            scheduled_at = datetime.strftime(scheduled_at, "%Y-%m-%d %H:%M:%S%z")
            click.echo(f"Toot scheduled for: {scheduled_at}")
        else:
            click.echo(f"Toot posted: {status['url']}")

    save_last_post_id(app, user, status["id"])

    delete_tmp_status_file()


@cli.command()
@click.argument("file", type=click.File(mode="rb"))
@click.option(
    "--description", "-d",
    help="Plain-text description of the media for accessibility purposes"
)
@json_option
@pass_context
def upload(
    ctx: Context,
    file: BinaryIO,
    description: Optional[str],
    json: bool,
):
    """Upload an image or video file

    This is probably not very useful, see `toot post --media` instead.
    """
    response = _do_upload(ctx.app, ctx.user, file, description, None)
    if json:
        click.echo(response.text)
    else:
        media = from_dict(MediaAttachment, response.json())
        click.echo()
        click.echo(f"Successfully uploaded media ID {media.id}, type '{media.type}'")
        click.echo(f"URL: {media.url}")
        click.echo(f"Preview URL: {media.preview_url}")


def _get_status_text(text, editor, media):
    isatty = sys.stdin.isatty()

    if not text and not isatty:
        text = sys.stdin.read().rstrip()

    if isatty:
        if editor:
            text = editor_input(editor, text)
        elif not text and not media:
            click.echo(f"Write or paste your toot. Press {EOF_KEY} to post it.")
            text = multiline_input()

    return text


def _get_scheduled_at(scheduled_at, scheduled_in):
    if scheduled_at:
        return scheduled_at

    if scheduled_in:
        scheduled_at = datetime.now(timezone.utc) + timedelta(seconds=scheduled_in)
        return scheduled_at.replace(microsecond=0).isoformat()

    return None


def _get_reply_to(app, user, reply_to, reply_last):
    if reply_last and reply_to:
        raise click.ClickException("--reply-last and --reply-to are mutually exclusive")

    if reply_last:
        last_id = get_last_post_id(app, user)
        if last_id:
            return last_id
        else:
            user = f"{user.username}@{app.instance}"
            raise click.ClickException(f"Cannot reply-last, no previous post ID found for {user}")

    return reply_to


def _upload_media(app, user, media, descriptions, thumbnails):
    # Match media to corresponding descriptions and thumbnail
    media = media or []
    descriptions = descriptions or []
    thumbnails = thumbnails or []
    uploaded_media = []

    for idx, file in enumerate(media):
        description = descriptions[idx].strip() if idx < len(descriptions) else None
        thumbnail = thumbnails[idx] if idx < len(thumbnails) else None
        result = _do_upload(app, user, file, description, thumbnail).json()
        uploaded_media.append(result)

    _wait_until_all_processed(app, user, uploaded_media)

    return [m["id"] for m in uploaded_media]


def _do_upload(app, user, file, description, thumbnail):
    return api.upload_media(app, user, file, description=description, thumbnail=thumbnail)


def _wait_until_all_processed(app, user, uploaded_media):
    """
    Media is uploaded asynchronously, and cannot be attached until the server
    has finished processing it. This function waits for that to happen.

    Once media is processed, it will have the URL populated.
    """
    if all(m["url"] for m in uploaded_media):
        return

    # Timeout after waiting 1 minute
    start_time = time()
    timeout = 60

    click.echo("Waiting for media to finish processing...")
    for media in uploaded_media:
        _wait_until_processed(app, user, media, start_time, timeout)


def _wait_until_processed(app, user, media, start_time, timeout):
    if media["url"]:
        return

    media = api.get_media(app, user, media["id"])
    while not media["url"]:
        sleep(1)
        if time() > start_time + timeout:
            raise click.ClickException(f"Media not processed by server after {timeout} seconds. Aborting.")
        media = api.get_media(app, user, media["id"])



import json as pyjson

import click

from toot import api
from toot.cli import Context, cli, json_option, pass_context
from toot.output import print_acct_list


@cli.group()
def follow_requests():
    """Manage follow requests"""
    pass


@follow_requests.command()
@click.argument("account")
@json_option
@pass_context
def accept(ctx: Context, account: str, json: bool):
    """Accept follow request from an account"""
    found_account = api.find_account(ctx.app, ctx.user, account)
    response = api.accept_follow_request(ctx.app, ctx.user, found_account["id"])
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ {account} is now following you", fg="green")


@follow_requests.command()
@click.argument("account")
@json_option
@pass_context
def reject(ctx: Context, account: str, json: bool):
    """Reject follow request from an account"""
    found_account = api.find_account(ctx.app, ctx.user, account)
    response = api.reject_follow_request(ctx.app, ctx.user, found_account["id"])
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ follow request from {account} rejected", fg="green")


@follow_requests.command()
@json_option
@pass_context
def list(ctx: Context, json: bool):
    """List follow requests"""
    requests = api.list_follow_requests(ctx.app, ctx.user)
    if json:
        click.echo(pyjson.dumps(requests))
    else:
        print_acct_list(requests)



from __future__ import annotations

__all__ = ("Hyperlink",)

from typing import Generator, List, Optional, Tuple

import urwid

from .text_embed import DisplayAttribute

# NOTE: Any new "private" attribute of any subclass of an urwid class should be
# prepended with "_uw" to avoid clashes with names used by urwid itself.


ESC = "\033"
OSC = f"{ESC}]"
ST = f"{ESC}\\"
START = f"{OSC}8;id=%d;%s{ST}".encode()
END = f"{OSC}8;;{ST}".encode()

valid_byte_range = range(32, 127)


class Hyperlink(urwid.WidgetWrap):
    """A widget containing hyperlinked text.

    Args:
        uri: The target of the hyperlink in URI (Uniform Resource Identifier)-encoded
          form.

          May be a web address (``http://...`` or ``https://...``), FTP address
          (``ftp://...``), local file (``file://...``), e-mail address (``mailto:``),
          etc.

          Every byte of this string, after being encoded, must be within the range
          ``32`` to ``126`` (both inclusive).

        attr: Display attribute of the hyperlink text.
        text: Alternative hyperlink text. If not given or ``None``, the URI itself is
          used. Must be a single-line string.

    Raises:
        TypeError: An argument is of an unexpected type.
        ValueError: An argument is of an expected type but of an unexpected value

    This widget always renders a **single line**, with *left* alignment and the
    *ellipsis* wrap mode of :py:class:`urwid.Text` i.e if the widget is rendered with
    a width less than the length of the hyperlink text, it is clipped at the right end
    with an ellipsis appended; on the other hand, if rendered with a width greater,
    it is padded with spaces on the right end.

    This widget is intended to be embedded in a :py:class:`~urwidgets.TextEmbed` widget
    to combine it with pure text or other widgets but may as well be used otherwise.

    This widget utilizes the ``OSC 8`` escape sequence implemented by a sizable number
    of mainstream terminal emulators. It utilizes the escape sequence in such a way that
    hyperlinks right next to one another should be detected, highlighted and treated as
    separate by any terminal emulator that correctly implements the feature. Also, if a
    hyperlink is wrapped or clipped, it shouldn't break.

    .. collapse:: Examples:

       >>> from urwidgets import Hyperlink
       >>>
       >>> url = "https://urwid.org"
       >>>
       >>> # The hyperlinks in the outputs should be highlighted on mouse hover
       >>> # and clickable (in the terminal), if supported.
       >>>
       >>> # Raw URI
       >>> link = Hyperlink(url)
       >>> canv = link.render(())
       >>> print(canv.text[0].decode())
       https://urwid.org
       >>>
       >>> # Clipped (with an ellipsis appended) when the render width (maxcols) is
       >>> # shorter than the link text
       >>> canv = link.render((len(url) - 4,))
       >>> print(canv.text[0].decode())
       https://urwid…
       >>>
       >>> # URI with custom text
       >>> hyperlink = Hyperlink(url, text="Urwid Website")
       >>> canv = hyperlink.render(())
       >>> print(canv.text[0].decode())
       Urwid Website

    .. seealso::

       `OSC 8 Specification \
       <https://gist.github.com/egmontkob/eb114294efbcd5adb1944c9f3cb5feda>`_
          Official specification for hyperlinks in terminals.

       `OSC 8 adoption in terminal emulators \
       <https://github.com/Alhadis/OSC8-Adoption>`_
          Documentation of the adoption of the feature across terminal emulators and
          terminal-based applications.

       :py:class:`~urwidgets.TextEmbed`
          A widget that enables the use of hyperlinks amidst normal text.
    """

    no_cache = ["render"]

    def __init__(
        self,
        uri: str,
        attr: DisplayAttribute = None,
        text: Optional[str] = None,
    ) -> None:
        self._uw_set_uri(uri)
        super().__init__(urwid.Text((_Attr(attr), ""), "left", "ellipsis"))
        self._uw_set_text(uri if text is None else text)

    def render(self, size: Tuple[int,], focus: bool = False) -> urwid.HyperlinkCanvas:
        return HyperlinkCanvas(self._uw_uri, self._w.render(size, focus))

    def _uw_set_text(self, text: str):
        if not isinstance(text, str):
            raise TypeError(f"Invalid type for 'text' (got: {type(text).__name__!r})")
        if not text:
            raise ValueError("Hyperlink text is empty")
        if "\n" in text:  # Other multi-line whitespace characters are escaped by urwid
            raise ValueError(f"Multi-line text (got: {text!r})")
        self._w.set_text((self._w.attrib[0][0], text))

    def _uw_set_uri(self, uri: str):
        if not isinstance(uri, str):
            raise TypeError(f"Invalid type for 'uri' (got: {type(uri).__name__!r})")
        if not uri:
            raise ValueError("URI is empty")
        invalid_bytes = frozenset(uri.encode()).difference(valid_byte_range)
        if invalid_bytes:
            raise ValueError(
                f"Invalid byte '\\x{tuple(invalid_bytes)[0]:02x}' found in URI: {uri!r}"
            )
        self._uw_uri = uri

    attrib = property(
        lambda self: self._w.attrib[0][0].attr,
        lambda self, attrib: self._w.set_text((_Attr(attrib), self._w.text)),
        doc="""The display attirbute of the hyperlink.

        :type: DisplayAttribute

        GET:
            Returns the display attirbute.

        SET:
            Sets the display attirbute.
        """,
    )

    text = property(
        lambda self: self._w.text,
        _uw_set_text,
        doc="""The alternate text of the hyperlink.

        :type: str

        GET:
            Returns the alternate text.

        SET:
            Sets the alternate text.
        """,
    )

    uri = property(
        lambda self: self._uw_uri,
        _uw_set_uri,
        doc="""The target of the hyperlink.

        :type: str

        GET:
            Returns the target.

        SET:
            Sets the target.
        """,
    )


class HyperlinkCanvas(urwid.Canvas):
    cacheable = False

    _uw_next_id = 0
    _uw_free_ids = set()

    def __init__(self, uri: str, text_canv: urwid.TextCanvas) -> None:
        super().__init__()
        self._uw_text_canv = text_canv
        self._uw_uri = uri.encode()
        self._uw_id = self._uw_get_id()

    def __del__(self):
        __class__._uw_free_ids.add(self._uw_id)

    def cols(self):
        return self._uw_text_canv.cols()

    def content(
        self,
        trim_left: int = 0,
        trim_top: int = 0,
        cols: int | None = None,
        rows: int | None = None,
        attr: DisplayAttribute = None,
    ) -> Generator[List[Tuple[DisplayAttribute, Optional[str], bytes]], None, None]:
        # There can only be one line since wrap="ellipsis" and the text was checked
        # to not contain "\n".
        content_line = next(
            self._uw_text_canv.content(trim_left, trim_top, cols, rows, attr)
        )

        if isinstance(content_line[0][0], _Attr):
            hyperlink_text, *padding = content_line
            link_attr = hyperlink_text[0].attr
            yield [
                (None, "U", START % (self._uw_id, self._uw_uri)),
                (
                    attr.get(link_attr, link_attr) if attr else link_attr,
                    *hyperlink_text[1:],
                ),
                (None, "U", END),
                *padding,  # if any
            ]
        else:  # A trim containing padding only
            yield content_line

    def rows(self):
        return self._uw_text_canv.rows()

    @staticmethod
    def _uw_get_id():
        if __class__._uw_free_ids:
            return __class__._uw_free_ids.pop()
        __class__._uw_next_id += 1

        return __class__._uw_next_id - 1


class _Attr:
    """Wraps a text display attribute to ensure it's always distinguished from those of
    neighbouring text runs.
    """

    def __init__(self, attr: DisplayAttribute):
        self.attr = attr



from typing import Tuple

import click

from toot import api
from toot.cli import Context, cli, json_option, pass_context
from toot.entities import Poll, from_response
from toot.output import print_poll


@cli.group()
def polls():
    """Show and vote on polls"""
    pass


@polls.command()
@click.argument("poll_id")
@json_option
@pass_context
def show(ctx: Context, poll_id: str, json: bool):
    """Show a poll by ID"""
    response = api.get_poll(ctx.app, ctx.user, poll_id)

    if json:
        click.echo(response.text)
    else:
        poll = from_response(Poll, response)
        print_poll(poll)


@polls.command()
@click.argument("poll_id")
@click.argument("choices", type=int, nargs=-1, required=True)
@json_option
@pass_context
def vote(ctx: Context, poll_id: str, choices: Tuple[int], json: bool):
    """Vote on a poll

    CHOICES is one or more zero-indexed integers corresponding to the desired poll choices.

    e.g. to vote for the first and third options use:

        toot polls vote <poll_id> 0 2
    """
    response = api.vote_poll(ctx.app, ctx.user, poll_id, choices)

    if json:
        click.echo(response.text)
    else:
        poll = from_response(Poll, response)
        click.echo("You voted!\n")
        print_poll(poll)



from __future__ import annotations

__all__ = (
    "parse_text",
    "TextEmbed",
    # Type Aliases
    "Markup",
    "StringMarkup",
    "ListMarkup",
    "TupleMarkup",
    "NormalTupleMarkup",
    "DisplayAttribute",
    "WidgetTupleMarkup",
    "WidgetListMarkup",
)

import re
from functools import lru_cache
from itertools import islice
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

import urwid

# NOTE: Any new "private" attribute of any subclass of an urwid class should be
# prepended with "_uw" to avoid clashes with names used by urwid itself.

# I really hope these are correct :D
Markup = Union["StringMarkup", "ListMarkup", "TupleMarkup"]
StringMarkup = Union[str, bytes]
ListMarkup = List["Markup"]
TupleMarkup = Union["NormalTupleMarkup", "WidgetTupleMarkup"]
NormalTupleMarkup = Tuple["DisplayAttribute", Union["StringMarkup", "ListMarkup"]]
DisplayAttribute = Union[None, str, bytes, "urwid.AttrSpec"]
WidgetTupleMarkup = Tuple[int, Union["urwid.Widget", "WidgetListMarkup"]]
WidgetListMarkup = List[Union["urwid.Widget", "Markup", "WidgetListMarkup"]]


class TextEmbed(urwid.Text):
    """A text widget within which other widgets may be embedded.

    This is an extension of the :py:class:`urwid.Text` widget. Every feature and
    interface of :py:class:`~urwid.Text` is supported and works essentially the same,
    **except for the "ellipsis" wrap mode** which is currently not implemented.
    Text markup format is essentially the same, except when embedding widgets.

    **Embedding Widgets**
        A widget is embedded by specifying it as a markup element with an **integer
        display attribute**, where the display attribute is the number of screen
        columns the widget should occupy.

        .. collapse:: Examples:

           >>> # w1 spans 2 columns
           >>> TextEmbed(["This widget (", (2, w1), ") spans two columns"])
           >>> # w1 and w2 span 2 columns
           >>> TextEmbed(["These widgets (", (2, [w1, w2]), ") span two columns each"])
           >>> # w1 and w2 span 2 columns, the text in-between has no display attribute
           >>> TextEmbed([(2, [w1, (None, "and"), w2]), " span two columns each"])
           >>> # w1 and w2 span 2 columns, text in the middle is red
           >>> TextEmbed((2, [w1, ("red", " i am red "), w2]))
           >>> # w1 and w3 span 2 columns, w2 spans 5 columns
           >>> TextEmbed((2, [w1, (5, w2), w3]))

        Visible embedded widgets are always rendered (may be cached) whenever the
        ``TextEmbed`` widget is re-rendered (i.e an uncached render). Hence, this
        allows for dynamic parts of text without updating the entire widget.
        Going a step further, embeddded widgets can be swapped using
        :py:class:`urwid.WidgetPlaceholder` but their widths will remain the same.

        .. note::

           - Every embedded widget must be a box widget and is always rendered with
             size ``(width, 1)``.  :py:class:`urwid.Filler` can be used to wrap flow
             widgets.
           - As regards the "space" wrap mode, each embedded widget is treated as a
             single WORD (i.e containing no whitespace). In other words, whitespace
             within embedded widgets do not influence wrapping.
           - After updating or swapping an embedded widget, the containing
             :py:class:`TextEmbed` widget's canvases should be invalidated to ensure
             it re-renders.

    Raises:
        TypeError: A widget markup element has a non-integer display attribute.
        ValueError: A widget doesn't support box sizing.
        ValueError: A widget has a non-positive width (display attribute).

    .. collapse:: Example:

        >>> from urwidgets import TextEmbed, Hyperlink
        >>> from urwid import Filler
        >>>
        >>> url = "https://urwid.org"
        >>> this = Hyperlink(url, text="This")
        >>> link = Hyperlink(url)
        >>>
        >>> text_embed = TextEmbed(
        ...     [
        ...         (4, Filler(this)),
        ...         " is a ",
        ...         ("bold", "link"),
        ...         " to ",
        ...         (len(url), Filler(link)),
        ...     ]
        ... )
        >>>
        >>> canv = text_embed.render(())
        >>> # The hyperlinks (`This` and `https://urwid.org`) should be highlighted
        >>> # on mouse hover and clickable (in the terminal), if supported.
        >>> print(canv.text[0].decode())
        This is a link to https://urwid.org

    .. seealso::

       :py:func:`parse_text`
          Parses a string into a text/widget markup that can be used with this class.
    """

    PLACEHOLDER_HEAD: ClassVar[str] = "\uf8fe"
    """"""  # Gets `autodoc` to include the member.

    PLACEHOLDER_TAIL: ClassVar[str] = "\uf8ff"
    """Embedded widgets' text placeholder components.

    Each should be a unique unicode codepoint that:

    - occupies exactly one column on a terminal screen.
    - is guaranteed to not occur in the text content of the widget, if any.

    Either or both may only be overriden on **subclasses (during their creation,
    not after)**, as in::

       class TextEmbedSub(TextEmbed):
           PLACEHOLDER_HEAD = "="
           PLACEHOLDER_TAIL = "-"

    NOTE:
        In most cases, the defaults should be sufficient. There's no need to override
        these except it's possible for the default values to occur in the widget's
        text content (if any), which is highly unlikely.

        That said, the default values should be considered implementation detail;
        hence, may change at any time without notice. They're only provided to help
        the user avoid conflicts with actual text content. If the values are depended
        upon, then they should be overriden on a subclass, as described above.
    """

    # In case a placeholder gets wrapped or clipped, this pattern will only match the
    # head of a placeholder not tails on subsequent lines
    _UW_PLACEHOLDER_PATTERN = re.compile(f"({PLACEHOLDER_HEAD}{PLACEHOLDER_TAIL}*)")

    # A tail must occur at the beginning of a line but may be preceded by padding
    # spaces when `align != "left"` and `wrap != "clip"`
    _UW_TAIL_PATTERN = re.compile(f"^( *)({PLACEHOLDER_TAIL}+)")

    def __init_subclass__(cls, **kwargs: Any) -> None:
        placeholder_tail_overriden = "PLACEHOLDER_TAIL" in cls.__dict__
        if "PLACEHOLDER_HEAD" in cls.__dict__ or placeholder_tail_overriden:
            cls._UW_PLACEHOLDER_PATTERN = re.compile(
                f"({cls.PLACEHOLDER_HEAD}{cls.PLACEHOLDER_TAIL}*)"
            )
        if placeholder_tail_overriden:
            cls._UW_TAIL_PATTERN = re.compile(f"^( *)({cls.PLACEHOLDER_TAIL}+)")

    attrib = property(
        lambda self: super().attrib,
        doc="""Run-length encoding of display attributes of the widget's content.

        :type: List[Tuple[Union[DisplayAttribute, int], int]]

        See the description of the second item in the return value of
        :py:meth:`get_text`.
        """,
    )

    embedded = property(
        lambda self: [(widget, width) for widget, width, _ in self._uw_embedded],
        doc="""Embedded widgets.

        Returns:
            A list of all embedded widgets and their respective widths, in the same
            order in which they were given in the text markup.

        :type: List[Tuple[urwid.Widget, int]]
        """,
    )

    text = property(
        lambda self: super().text,
        doc="""Raw text content of the widget.

        :type: str

        See the description of the first item in the return value of
        :py:meth:`get_text`.
        """,
    )

    def get_text(
        self,
    ) -> Tuple[str, List[Tuple[Union[DisplayAttribute, int], int]]]:
        """Returns a representation of the widget's content.

        Returns:
            A tuple ``(text, attrib)``, where

            - *text* is the raw text content of the widget.

              Each embedded widget is represented by a placeholder substring
              with length equal to the widget's width.

            - *attrib* is the run-length encoding of display attributes.

              Any entry containing a display attribute of the ``int`` type (e.g
              ``(1, 4)``) denotes an embedded widget, where the display attirbute is
              the index of the widget within the :py:attr:`embedded` widgets list and
              the run length is the width of the widget.
        """
        return super().get_text()

    def render(
        self, size: Tuple[int,], focus: bool = False
    ) -> Union[urwid.TextCanvas, urwid.CompositeCanvas]:
        text_canv = fix_text_canvas_attr(super().render(size, focus))
        embedded = self._uw_embedded
        if not embedded:
            return text_canv

        def append_text_lines():
            nonlocal top

            if n_lines:
                partial_canv = urwid.CompositeCanvas(text_canv)
                partial_canv.trim(top, n_lines)
                canvases.append((partial_canv, None, focus))
                top += n_lines

        text = text_canv.text
        canvases = []
        placeholder_pattern = type(self)._UW_PLACEHOLDER_PATTERN
        placeholder_tail = type(self).PLACEHOLDER_TAIL
        tail = None
        top = 0
        n_lines = 0
        clipped = self.wrap == "clip"

        if clipped:
            if self.align != "left":
                translation = self.get_line_translation(size[0])
            text_canv_content = tuple(text_canv.content())
        else:
            embedded_iter = iter(embedded)

        for row_index, line in enumerate(text):
            line = line.decode()
            if clipped:
                if line.startswith(placeholder_tail):  # align != "left"
                    widget_index = text_canv_content[row_index][0][0]
                    widget, width, start_pos = embedded[widget_index]
                    tail_canv = widget.render((width, 1), focus)
                    left_trim = -translation[row_index][0][0]
                    # the placeholder is clipped => left_trim > start_pos
                    tail_width = width - (left_trim - start_pos)
                    tail = (tail_width, tail_canv)
                    embedded_iter = islice(embedded, widget_index + 1, None)
                else:
                    tail = None
            if tail:
                if clipped:
                    append_text_lines()
                line_canv = urwid.CompositeCanvas(text_canv)
                line_canv.trim(top, 1)
                partial_canv, tail = self._uw_embed(
                    line, line_canv, embedded_iter, focus, tail
                )
                canvases.append((partial_canv, None, focus))
                n_lines = 0
                top += 1
            elif placeholder_pattern.search(line):
                append_text_lines()
                if clipped:
                    for attr, *_ in text_canv_content[row_index]:
                        if isinstance(attr, int):
                            break
                    embedded_iter = islice(embedded, attr, None)
                line_canv = urwid.CompositeCanvas(text_canv)
                line_canv.trim(top, 1)
                partial_canv, tail = self._uw_embed(
                    line, line_canv, embedded_iter, focus
                )
                canvases.append((partial_canv, None, focus))
                n_lines = 0
                top += 1
            else:
                n_lines += 1
        append_text_lines()

        return urwid.CanvasCombine(canvases)

    def set_text(self, markup: Markup) -> None:
        """Sets the widget's content.

        Also supports widget markup elements. See the class description.
        """
        markup, self._uw_embedded = self._uw_substitute_widgets(markup)
        super().set_text(markup)
        self._uw_update_widget_start_pos()

    def set_wrap_mode(self, mode: str) -> None:
        if mode == "ellipsis":
            raise NotImplementedError("Wrap mode 'ellipsis' is not implemented.")
        super().set_wrap_mode(mode)

    wrap = property(lambda self: super().wrap, set_wrap_mode)

    def _uw_update_widget_start_pos(self) -> None:
        """Updates the start position of embedded widgets on their respective lines."""
        if not self._uw_embedded:
            return

        # - Text is clipped per line.
        # - Since the pad/trim amount in the translation (produced by
        #   `StandardTextLayout.align_layout()`) is relative to the start of the line
        #   wrt the layout width (maxcol), the position of an embedded widgets on its
        #   respective line should be relative to the start of the line, not considering
        #   alignment.
        find_placeholders = type(self)._UW_PLACEHOLDER_PATTERN.finditer
        embedded_iter = iter(self._uw_embedded)
        self._uw_embedded = [
            # Using `calc_width()` instead of `match.start()` directly to account for
            # wide and zero-width characters
            (widget, width, urwid.calc_width(line, 0, match.start()))
            for line in super().get_text()[0].splitlines()
            for match, (widget, width, _) in zip(find_placeholders(line), embedded_iter)
        ]

    @classmethod
    def _uw_substitute_widgets(
        cls, markup: Markup
    ) -> Tuple[Markup, List[Tuple[urwid.Widget, int, int]]]:
        """Extracts embedded widgets from *markup* and replace widget markup elements
        with placeholders.

        Returns:
            A tuple containing:

            - The given markup flattened and with all widget elements replaced by
              placeholders.
            - A list of ``(widget, width, start_position)`` tuples describing the
              embedded widgets, where *start_position* is initialized to zero and
              later updated by :py:meth:`_uw_update_widget_start_pos`.
        """

        def recurse_markup(attr: Union[DisplayAttribute, int], markup: Markup) -> None:
            if isinstance(markup, list):
                for markup in markup:
                    recurse_markup(attr, markup)
            elif isinstance(markup, tuple):
                if len(markup) != 2:
                    raise urwid.TagMarkupException(
                        "Tuples must be in the form `(attribute, tagmarkup)` "
                        f"(got: {markup!r})"
                    )
                recurse_markup(*markup)
            elif isinstance(markup, urwid.Widget):
                if not isinstance(attr, int):
                    raise TypeError(
                        "Invalid type for embedded widget width "
                        f"(got: {type(attr).__name__!r})"
                    )
                if "box" not in markup.sizing():
                    raise ValueError(f"Not a box widget (got: {markup!r})")
                if attr <= 0:
                    raise ValueError(f"Invalid widget width (got: {attr!r})")
                new_markup.append(
                    (len(embedded), placeholder_head + placeholder_tail * (attr - 1))
                )
                embedded.append((markup, attr, 0))
            else:
                # Normalize text type to `str` since other parts of this class use
                # and expect `str`
                if isinstance(markup, bytes):
                    markup = markup.decode()
                new_markup.append(markup if attr is None else (attr, markup))

        embedded = []
        new_markup = []
        placeholder_head = cls.PLACEHOLDER_HEAD
        placeholder_tail = cls.PLACEHOLDER_TAIL
        recurse_markup(None, markup)

        return new_markup, embedded

    @classmethod
    def _uw_embed(
        cls,
        line: str,
        line_canv: urwid.CompositeCanvas,
        embedded_iter: Iterator[Tuple[urwid.Widget, int, int]],
        focus: bool = False,
        tail: Optional[Tuple[int, urwid.Canvas]] = None,
    ) -> Tuple[urwid.CompositeCanvas, Optional[Tuple[int, urwid.Canvas]]]:
        """Replaces widget placeholders in a line with with the widgets' contents.

        Args:
            line: A line of the original text canvas.
            line_canv: A canvas corresponding to *line*.
            embedded_iter: An iterator of ``(widget, width, start_position)`` tuples
              in the same order as :py:attr:`embedded`, where *start_position* is as
              determined by :py:meth:`_uw_update_widget_start_pos`.
            focus: As in :py:meth:`render`.
            tail: The description of the "tail" of an embedded widget that is the first
              part of the line ``(tail_width, tail_canv)``, if it was wrapped/clipped,
              where:

              - *tail_width* is the width of the remaining (unused) portion of the
                widget's canvas content towards it's right end.
              - *tail_canv* is the original rendered canvas of the widget, unmodified.

              OR ``None`` if a widget is not the first part of the line.

        Returns:
            A tuple containing:

            - A ``CompositeCanvas`` containing the separate parts from the original
              text canvas and the embedded widgets' canvases.
            - The description of the "tail" of an embedded widget that is the last part
              of the line ``(tail_width, tail_canv)`` (see the description of *tail*
              above), if it was wrapped/clipped OR ``None`` if it wasn't wrapped/clipped
              or a widget is not the last part of the line.
        """
        canvases = []
        line_index = 0

        if tail:
            # - Since this is the line after the head, then it must contain [a part of]
            #   the tail
            # - Only one possible occurence of a tail per line
            # - Might be preceded by padding spaces when `align != "left"`
            _, padding, tail_string, line = cls._UW_TAIL_PATTERN.split(line)

            if padding:
                # Can use `len(padding)` since all characters should be spaces
                canv = urwid.Text(padding).render((len(padding),), focus)
                canvases.append((canv, None, focus, len(padding)))
                line_index += len(padding)

            tail_width, tail_canv = tail
            canv = urwid.CompositeCanvas(tail_canv)
            canv.pad_trim_left_right(tail_width - tail_canv.cols(), 0)
            canvases.append((canv, None, focus, len(tail_string)))
            line_index += len(tail_string)

            if not line:
                tail = (
                    (tail_width - len(tail_string), tail_canv)
                    if len(tail_string) < tail_width
                    else None
                )
                return urwid.CanvasJoin(canvases), tail
            tail = None

        placeholder_pattern = cls._UW_PLACEHOLDER_PATTERN

        for part in placeholder_pattern.split(line):
            if not part:
                continue

            if placeholder_pattern.fullmatch(part):
                widget, width, _ = next(embedded_iter)
                canv = widget.render((width, 1), focus)
                # `len(part)`, in case the placeholder was wrapped
                canvases.append((canv, None, focus, len(part)))
                line_index += len(part)
                if len(part) != width:
                    tail = (width - len(part), canv)
            else:
                # Using `calc_width()` instead of `len(part)` directly to account for
                # wide and zero-width characters
                maxcol = urwid.calc_width(part, 0, len(part))
                canv = urwid.CompositeCanvas(line_canv)
                canv.pad_trim_left_right(-line_index, 0)
                canvases.append((canv, None, focus, maxcol))
                line_index += maxcol

        return urwid.CanvasJoin(canvases), tail


def parse_text(
    text: str,
    patterns: Iterable[re.Pattern],
    repl: Callable[[re.Pattern, Tuple[Optional[str]], Tuple[int, int], ...], Markup],
    *repl_args: Any,
    **repl_kwargs: Any,
) -> Markup:
    r"""Parses a string into a text/widget markup.

    Args:
        text: The string to parse.
        patterns: An iterable of RegEx pattern objects.
        repl: A callable to replace a substring of *text* matched by any of the given
          RegEx patterns.
        repl_args: Additional positional arguments to be passed to *repl* whenever it's
          called.
        repl_kwargs: keyword arguments to be passed to *repl* whenever it's called.

    Returns:
        A text/widget markup (see :py:data:`Markup`) that should be compatible with
        :py:class:`TextEmbed` and/or :py:class:`urwid.Text`, depending on the values
        returned by *repl*.

    Raises:
        TypeError: An argument is of an unexpected type.
        ValueError: *patterns* is empty.
        ValueError: A given pattern object was not compiled from a :py:class:`str`
          instance.

    Whenever any of the given RegEx patterns matches a **non-empty** substring of
    *text*, *repl* is called with the following arguments (in the given order):

    - the :py:class:`~re.Pattern` object that matched the substring
    - a tuple containing the match groups

      - starting with the whole match,
      - followed by the all the subgroups of the match, from 1 up to however many
        groups are in the pattern, if any (``None`` for each group that didn't
        participate in the match)

    - a tuple containing the span (start and end indexes) of the substring
    - *repl_args* unpacked
    - *repl_kwargs* unpacked

    and *should* return a valid text/widget markup (see :py:data:`Markup`). If the
    value returned is *false* (such as ``None`` or an empty string), it is omitted
    from the result.

    .. collapse:: Example:

        >>> import re
        >>> from urwid import Filler
        >>> from urwidgets import Hyperlink, TextEmbed, parse_text
        >>>
        >>> MARKDOWN = {
        >>>     re.compile(r"\*\*(.+?)\*\*"): lambda g: ("bold", g[1]),
        >>>     re.compile("https://[^ ]+"): (
        >>>         lambda g: (len(g[0]), Filler(Hyperlink(g[0])))
        >>>     ),
        >>>     re.compile(r"\[(.+)\]\((.+)\)"): (
        >>>         lambda g: (len(g[1]), Filler(Hyperlink(g[2], text=g[1])))
        >>>     ),
        >>> }
        >>>
        >>> link = "https://urwid.org"
        >>> text = f"[This]({link}) is a **link** to {link}"
        >>> print(text)
        [This](https://urwid.org) is a **link** to https://urwid.org
        >>>
        >>> markup = parse_text(
        >>>     text, MARKDOWN, lambda pattern, groups, span: MARKDOWN[pattern](groups)
        >>> )
        >>> print(markup)
        [
          (4, <Filler box widget <Hyperlink flow widget>>),
          ' is a ',
          ('bold', 'link'),
          ' to ',
          (17, <Filler box widget <Hyperlink flow widget>>),
        ]
        >>>
        >>> text_widget = TextEmbed(markup)
        >>> canv = text_widget.render(())
        >>> # The hyperlinks (`This` and `https://urwid.org`) should be highlighted
        >>> # on mouse hover and clickable (in the terminal), if supported.
        >>> print(canv.text[0].decode())
        This is a link to https://urwid.org

    NOTE:
        In the case of overlapping matches, the substring that occurs first is matched
        and if they start at the same index, the pattern that appears first in
        *patterns* takes precedence.
    """
    if not isinstance(text, str):
        raise TypeError(f"Tnvalid type for 'text' (got: {type(text).__name__!r})")
    if not text:
        return text

    patterns = tuple(patterns)
    if not patterns:
        raise ValueError("No RegEx patterns")

    combined_pattern, indexed_patterns = combine_patterns(patterns)
    full_markup = []
    ptr = 0
    for match in combined_pattern.finditer(text):
        span = match.span()
        if ptr < span[0]:
            full_markup.append(text[ptr : span[0]])
        if match.group():
            pattern_index = match.lastindex
            pattern = indexed_patterns[pattern_index]
            markup = repl(
                pattern,
                match.groups()[pattern_index - 1 : pattern_index + pattern.groups],
                span,
                *repl_args,
                **repl_kwargs,
            )
            if markup:
                full_markup.append(markup)
        ptr = span[1]
    if ptr < len(text):
        full_markup.append(text[ptr:])

    return full_markup[0] if len(full_markup) == 1 else full_markup


# Private

RE_INLINE_FLAGS = {re.A: "a", re.I: "i", re.L: "L", re.M: "m", re.S: "s", re.X: "x"}


@lru_cache()
def combine_patterns(
    patterns: Tuple[re.Pattern],
) -> Tuple[re.Pattern, Dict[int, re.Pattern]]:
    """Combines multiple RegEx patterns with their respective flags into a single OR-ed
    pattern.

    Returns:
        A tuple containing

        - the combined RegEx pattern
        - a dictionary mapping the index of the group in the combined pattern
          corresponding to each given pattern to the pattern
    """
    grouped_patterns = []
    indexed_patterns = {}  # <index of group in combined pattern>: <pattern>
    group_index = 1
    for pattern in patterns:
        pattern_string = pattern.pattern
        if not isinstance(pattern_string, str):
            raise ValueError(f"Pattern not compiled from `str` (got: {pattern!r})")

        inline_flags = get_inline_flags(pattern.flags)
        grouped_patterns.append(
            f"(?{inline_flags}:({pattern_string}))"
            if inline_flags
            else f"({pattern_string})"
        )
        indexed_patterns[group_index] = pattern
        group_index += pattern.groups + 1

    return re.compile("|".join(grouped_patterns)), indexed_patterns


def fix_text_canvas_attr(canv: urwid.TextCanvas) -> urwid.TextCanvas:
    """Workaround for a bug in in `urwid.text_layout.StandardTextLayout`.

    When `wrap=clip, align=center` and there's a line starting with a markup that has
    a display attribute, when the render width (maxcol) is one less than the line's
    width (in screen columns, not characters), the line is rendered as an empty
    string.

    See https://github.com/urwid/urwid/issues/542.
    """
    for line_attr in canv._attr:
        if line_attr[0] == (None, 0):
            del line_attr[0]

    return canv


# Only 511 (zero is excluded) unique bit patterns (and not even all can occur)
@lru_cache(maxsize=None)
def get_inline_flags(flags: int) -> str:
    """Converts a RegEx integer flag into the corresponding set of inline flags"""
    return "".join([inline for flag, inline in RE_INLINE_FLAGS.items() if flag & flags])



import urwid

from toot import api
from toot.exceptions import ApiError
from toot.utils.datetime import parse_datetime
from .widgets import Button, CheckBox, RadioButton, RoundedLineBox
from .richtext import html_to_widgets


class Poll(urwid.ListBox):
    """View and vote on a poll"""

    def __init__(self, app, user, status):
        self.status = status
        self.app = app
        self.user = user
        self.poll = status.original.data.get("poll")
        self.button_group = []
        self.api_exception = None
        self.setup_listbox()

    def setup_listbox(self):
        actions = list(self.generate_contents(self.status))
        walker = urwid.SimpleListWalker(actions)
        super().__init__(walker)

    def build_linebox(self, contents):
        contents = urwid.Pile(list(contents))
        contents = urwid.Padding(contents, left=1, right=1)
        return RoundedLineBox(contents)

    def vote(self, button_widget):
        poll = self.status.original.data.get("poll")
        choices = []
        for idx, button in enumerate(self.button_group):
            if button.get_state():
                choices.append(idx)

        if len(choices):
            try:
                response = api.vote(self.app, self.user, poll["id"], choices=choices)
                self.status.original.data["poll"] = response
                self.api_exception = None
                self.poll["voted"] = True
                self.poll["own_votes"] = choices
            except ApiError as exception:
                self.api_exception = exception
            finally:
                self.setup_listbox()

    def generate_poll_detail(self):
        poll = self.poll

        self.button_group = []  # button group
        for idx, option in enumerate(poll["options"]):
            voted_for = (
                poll["voted"] and poll["own_votes"] and idx in poll["own_votes"]
            )

            if poll["voted"] or poll["expired"]:
                prefix = " ✓  " if voted_for else "    "
                yield urwid.Text(("dim", prefix + f'{option["title"]}'))
            else:
                if poll["multiple"]:
                    checkbox = CheckBox(f'{option["title"]}')
                    self.button_group.append(checkbox)
                    yield checkbox
                else:
                    yield RadioButton(self.button_group, f'{option["title"]}')

        yield urwid.Divider()

        poll_detail = "Poll · {} votes".format(poll["votes_count"])

        if poll["expired"]:
            poll_detail += " · Closed"

        if poll["expires_at"]:
            expires_at = parse_datetime(poll["expires_at"]).strftime(
                "%Y-%m-%d %H:%M"
            )
            poll_detail += " · Closes on {}".format(expires_at)

        yield urwid.Text(("dim", poll_detail))

    def generate_contents(self, status):
        yield urwid.Divider()

        widgetlist = html_to_widgets(status.data["content"])

        for line in widgetlist:
            yield (line)

        yield urwid.Divider()
        yield self.build_linebox(self.generate_poll_detail())
        yield urwid.Divider()

        if self.poll["voted"]:
            yield urwid.Text(("grey", "< Already Voted >"))
        elif not self.poll["expired"]:
            yield Button("Vote", on_press=self.vote)

        if self.api_exception:
            yield urwid.Divider()
            yield urwid.Text("warning", str(self.api_exception))



from copy import copy
import click

from toot import api
from toot.cli import cli, json_option, Context, pass_context
from toot.cli import VISIBILITY_CHOICES
from toot.output import print_table


@cli.command()
@click.argument("status_id")
@json_option
@pass_context
def delete(ctx: Context, status_id: str, json: bool):
    """Delete a status"""
    response = api.delete_status(ctx.app, ctx.user, status_id)
    if json:
        click.echo(response.text)
    else:
        click.secho("✓ Status deleted", fg="green")


@cli.command()
@click.argument("status_id")
@json_option
@pass_context
def favourite(ctx: Context, status_id: str, json: bool):
    """Favourite a status"""
    response = api.favourite(ctx.app, ctx.user, status_id)
    if json:
        click.echo(response.text)
    else:
        click.secho("✓ Status favourited", fg="green")


@cli.command()
@click.argument("status_id")
@json_option
@pass_context
def unfavourite(ctx: Context, status_id: str, json: bool):
    """Unfavourite a status"""
    response = api.unfavourite(ctx.app, ctx.user, status_id)
    if json:
        click.echo(response.text)
    else:
        click.secho("✓ Status unfavourited", fg="green")


@cli.command()
@click.argument("status_id")
@click.option(
    "--visibility", "-v",
    help="Post visibility",
    type=click.Choice(VISIBILITY_CHOICES),
    default="public",
)
@json_option
@pass_context
def reblog(ctx: Context, status_id: str, visibility: str, json: bool):
    """Reblog (boost) a status"""
    response = api.reblog(ctx.app, ctx.user, status_id, visibility=visibility)
    if json:
        click.echo(response.text)
    else:
        click.secho("✓ Status reblogged", fg="green")


@cli.command()
@click.argument("status_id")
@json_option
@pass_context
def unreblog(ctx: Context, status_id: str, json: bool):
    """Unreblog (unboost) a status"""
    response = api.unreblog(ctx.app, ctx.user, status_id)
    if json:
        click.echo(response.text)
    else:
        click.secho("✓ Status unreblogged", fg="green")


@cli.command()
@click.argument("status_id")
@json_option
@pass_context
def pin(ctx: Context, status_id: str, json: bool):
    """Pin a status"""
    response = api.pin(ctx.app, ctx.user, status_id)
    if json:
        click.echo(response.text)
    else:
        click.secho("✓ Status pinned", fg="green")


@cli.command()
@click.argument("status_id")
@json_option
@pass_context
def unpin(ctx: Context, status_id: str, json: bool):
    """Unpin a status"""
    response = api.unpin(ctx.app, ctx.user, status_id)
    if json:
        click.echo(response.text)
    else:
        click.secho("✓ Status unpinned", fg="green")


@cli.command()
@click.argument("status_id")
@json_option
@pass_context
def bookmark(ctx: Context, status_id: str, json: bool):
    """Bookmark a status"""
    response = api.bookmark(ctx.app, ctx.user, status_id)
    if json:
        click.echo(response.text)
    else:
        click.secho("✓ Status bookmarked", fg="green")


@cli.command()
@click.argument("status_id")
@json_option
@pass_context
def unbookmark(ctx: Context, status_id: str, json: bool):
    """Unbookmark a status"""
    response = api.unbookmark(ctx.app, ctx.user, status_id)
    if json:
        click.echo(response.text)
    else:
        click.secho("✓ Status unbookmarked", fg="green")


@cli.command()
@click.argument("status_id")
@json_option
@pass_context
def reblogged_by(ctx: Context, status_id: str, json: bool):
    """Show accounts that reblogged a status"""
    response = api.reblogged_by(ctx.app, ctx.user, status_id)

    if json:
        click.echo(response.text)
    else:
        rows = [[a["acct"], a["display_name"]] for a in response.json()]
        if rows:
            headers = ["Account", "Display name"]
            print_table(headers, rows)
        else:
            click.echo("This status is not reblogged by anyone")


# Make alias in snake case to keep BC
reblogged_by_alias = copy(reblogged_by)
reblogged_by_alias.hidden = True
cli.add_command(reblogged_by_alias, name="reblogged_by")



import sys
import click

from toot import api
from toot.cli import NOTIFICATION_TYPE_CHOICES, InstanceParamType, cli, get_context, pass_context, Context, json_option
from typing import Optional, Tuple
from toot.cli.validators import validate_instance

from toot.entities import Notification, Status, from_dict
from toot.output import print_notifications, print_timeline, print_warning


@cli.command()
@click.option(
    "--instance", "-i",
    type=InstanceParamType(),
    callback=validate_instance,
    help="""Domain or base URL of the instance from which to read,
            e.g. 'mastodon.social' or 'https://mastodon.social'""",
)
@click.option("--account", "-a", help="Show account timeline")
@click.option("--list", help="Show list timeline")
@click.option("--tag", "-t", help="Show hashtag timeline")
@click.option("--public", "-p", is_flag=True, help="Show public timeline")
@click.option(
    "--local", "-l", is_flag=True,
    help="Show only statuses from the local instance (public and tag timelines only)"
)
@click.option(
    "--reverse", "-r", is_flag=True,
    help="Reverse the order of the shown timeline (new posts at the bottom)"
)
@click.option(
    "--once", "-1", is_flag=True,
    help="Only show the first <count> toots, do not prompt to continue"
)
@click.option(
    "--count", "-c", type=int, default=10,
    help="Number of posts per page (max 20)"
)
def timeline(
    instance: Optional[str],
    account: Optional[str],
    list: Optional[str],
    tag: Optional[str],
    public: bool,
    local: bool,
    reverse: bool,
    once: bool,
    count: int,
):
    """Show timelines (deprecated, use `toot timelines` instead)

    By default shows the home timeline.
    """
    if len([arg for arg in [tag, list, public, account] if arg]) > 1:
        raise click.ClickException("Only one of --public, --tag, --account, or --list can be used at one time.")

    if local and not (public or tag):
        raise click.ClickException("The --local option is only valid alongside --public or --tag.")

    if instance and not (public or tag):
        raise click.ClickException("The --instance option is only valid alongside --public or --tag.")

    if public and instance:
        generator = api.anon_public_timeline_generator(instance, local, count)
    elif tag and instance:
        generator = api.anon_tag_timeline_generator(instance, tag, local, count)
    else:
        ctx = get_context()
        list_id = _get_list_id(ctx, list)

        """Show recent statuses in a timeline"""
        generator = api.get_timeline_generator(
            ctx.app,
            ctx.user,
            account=account,
            list_id=list_id,
            tag=tag,
            public=public,
            local=local,
            limit=count,
        )

    _show_timeline(generator, reverse, once)


@cli.command()
@click.option(
    "--reverse", "-r", is_flag=True,
    help="Reverse the order of the shown timeline (new posts at the bottom)"
)
@click.option(
    "--once", "-1", is_flag=True,
    help="Only show the first <count> toots, do not prompt to continue"
)
@click.option(
    "--count", "-c", type=int, default=10,
    help="Number of posts per page (max 20)"
)
@pass_context
def bookmarks(
    ctx: Context,
    reverse: bool,
    once: bool,
    count: int,
):
    """Show recent statuses in a timeline"""
    generator = api.bookmark_timeline_generator(ctx.app, ctx.user, limit=count)
    _show_timeline(generator, reverse, once)


@cli.command()
@click.option(
    "--clear", is_flag=True,
    help="Dismiss all notifications and exit"
)
@click.option(
    "--reverse", "-r", is_flag=True,
    help="Reverse the order of the shown notifications (newest on top)"
)
@click.option(
    "--type", "-t", "types",
    type=click.Choice(NOTIFICATION_TYPE_CHOICES),
    multiple=True,
    help="Types to include in the result, can be specified multiple times"
)
@click.option(
    "--exclude-type", "-e", "exclude_types",
    type=click.Choice(NOTIFICATION_TYPE_CHOICES),
    multiple=True,
    help="Types to exclude in the result, can be specified multiple times"
)
@click.option(
    "--mentions", "-m", is_flag=True,
    help="Show only mentions (same as --type mention, overrides --type, DEPRECATED)"
)
@json_option
@pass_context
def notifications(
    ctx: Context,
    clear: bool,
    reverse: bool,
    mentions: bool,
    types: Tuple[str],
    exclude_types: Tuple[str],
    json: bool,
):
    """Show notifications"""
    if clear:
        api.clear_notifications(ctx.app, ctx.user)
        click.secho("✓ Notifications cleared", fg="green")
        return

    if mentions:
        print_warning("`--mentions` option is deprecated in favour of `--type mentions`")
        types = ("mention",)

    response = api.get_notifications(ctx.app, ctx.user, types=types, exclude_types=exclude_types)

    if json:
        if reverse:
            print_warning("--reverse is not supported alongside --json, ignoring")
        click.echo(response.text)
        return

    notifications = [from_dict(Notification, n) for n in response.json()]
    if reverse:
        notifications = reversed(notifications)

    if notifications:
        print_notifications(notifications)
    else:
        click.echo("You have no notifications")


def _show_timeline(generator, reverse, once):
    while True:
        try:
            items = next(generator)
        except StopIteration:
            click.echo("That's all folks.")
            return

        if reverse:
            items = reversed(items)

        statuses = [from_dict(Status, item) for item in items]
        print_timeline(statuses)

        if once or not sys.stdout.isatty():
            break

        char = input("\nContinue? [Y/n] ")
        if char.lower() == "n":
            break


def _get_list_id(ctx: Context, value: Optional[str]) -> Optional[str]:
    if not value:
        return None

    lists = api.get_lists(ctx.app, ctx.user)
    for list in lists:
        if list["id"] == value or list["title"] == value:
            return list["id"]



import re
import shutil
import textwrap
import typing as t

import click
from wcwidth import wcswidth

from toot.entities import Account, Instance, List, Notification, Poll, Status
from toot.utils import get_text, html_to_paragraphs
from toot.wcstring import wc_wrap

DEFAULT_WIDTH = 80


def get_max_width() -> int:
    return click.get_current_context().max_content_width or DEFAULT_WIDTH


def get_terminal_width() -> int:
    return shutil.get_terminal_size().columns


def get_terminal_height() -> int:
    return shutil.get_terminal_size().lines


def get_width() -> int:
    return min(get_terminal_width(), get_max_width())


def print_warning(text: str):
    click.secho(f"Warning: {text}", fg="yellow", err=True)


def get_continue():
    click.secho(
        f"Press {green('Space')} or {green('Enter')} to continue, {yellow('Esc')} or {yellow('q')} to break."
    )
    while True:
        char = click.getchar()
        if char == " " or char == "\r":
            return True
        if char == "\x1b" or char == "q":
            return False


def print_instance(instance: Instance):
    for line in instance_lines(instance, get_width()):
        click.echo(line)


def instance_lines(instance: Instance, width: int) -> t.Generator[str, None, None]:
    yield f"{green(instance.title)}"
    yield f"{blue(instance.uri)}"
    yield f"running Mastodon {instance.version}"
    yield ""

    if instance.description:
        for paragraph in re.split(r"[\r\n]+", instance.description.strip()):
            paragraph = get_text(paragraph)
            yield textwrap.fill(paragraph, width=width)
            yield ""

    if instance.rules:
        yield "Rules:"
        for ordinal, rule in enumerate(instance.rules):
            ordinal = f"{ordinal + 1}."
            lines = textwrap.wrap(rule.text, width - len(ordinal))
            first = True
            for line in lines:
                if first:
                    yield f"{ordinal} {line}"
                    first = False
                else:
                    yield f"{' ' * len(ordinal)} {line}"
        yield ""

    contact = instance.contact_account
    if contact:
        yield f"Contact: {contact.display_name} @{contact.acct}"


def print_account(account: Account) -> None:
    width = get_width()
    click.echo(account_to_text(account, width))


def account_to_text(account: Account, width: int) -> str:
    return "\n".join(account_lines(account, width))


def account_lines(account: Account, width: int) -> t.Generator[str, None, None]:
    acct = f"@{account.acct}"
    since = account.created_at.strftime("%Y-%m-%d")

    yield f"{green(acct)} {account.display_name}"

    if account.note:
        yield ""
        yield from html_lines(account.note, width)

    yield ""
    yield f"ID: {green(account.id)}"
    yield f"Since: {green(since)}"
    yield ""
    yield f"Followers: {yellow(account.followers_count)}"
    yield f"Following: {yellow(account.following_count)}"
    yield f"Statuses: {yellow(account.statuses_count)}"

    if account.fields:
        for field in account.fields:
            name = field.name.title()
            yield f"\n{yellow(name)}:"
            yield from html_lines(field.value, width)
            if field.verified_at:
                yield green("✓ Verified")

    yield ""
    yield account.url


def print_acct_list(accounts):
    for account in accounts:
        acct = green(f"@{account['acct']}")
        click.echo(f"* {acct} {account['display_name']}")


def print_tag_list(tags):
    for tag in tags:
        click.echo(f"* {format_tag_name(tag)}\t{tag['url']}")


def print_lists(lists: t.List[List]):
    headers = ["ID", "Title", "Replies"]
    data = [[lst.id, lst.title, lst.replies_policy or ""] for lst in lists]
    print_table(headers, data)


def print_table(headers: t.List[str], data: t.List[t.List[str]]):
    widths = [[len(cell) for cell in row] for row in data + [headers]]
    widths = [max(width) for width in zip(*widths)]

    def print_row(row):
        for idx, cell in enumerate(row):
            width = widths[idx]
            click.echo(cell.ljust(width), nl=False)
            click.echo("  ", nl=False)
        click.echo()

    underlines = ["-" * width for width in widths]

    print_row(headers)
    print_row(underlines)

    for row in data:
        print_row(row)


def print_list_accounts(accounts):
    if accounts:
        click.echo("Accounts in list:\n")
        print_acct_list(accounts)
    else:
        click.echo("This list has no accounts.")


def print_search_results(results):
    accounts = results.get("accounts")
    hashtags = results.get("hashtags")
    statuses = results.get("statuses")

    if accounts:
        click.echo("\nAccounts:")
        print_acct_list(accounts)

    if hashtags:
        click.echo("\nHashtags:")
        click.echo(", ".join([format_tag_name(tag) for tag in hashtags]))

    if statuses:
        click.echo("\nStatuses:")
        for status in statuses:
            click.echo(f" * {green(status['id'])} {status['url']}")

    if not accounts and not hashtags and not statuses:
        click.echo("Nothing found")


def print_status(status: Status) -> None:
    for line in status_lines(status, get_width()):
        click.echo(line)


def status_lines(status: Status, width: int) -> t.Generator[str, None, None]:
    width = get_width()
    status_id = status.id
    in_reply_to_id = status.in_reply_to_id
    reblogged_by = status.account if status.reblog else None
    status = status.original

    time = status.created_at.strftime("%Y-%m-%d %H:%M %Z")
    username = "@" + status.account.acct
    spacing = width - wcswidth(username) - wcswidth(time) - 2

    display_name = status.account.display_name

    if display_name:
        author = f"{green(display_name)} {blue(username)}"
        spacing -= wcswidth(display_name) + 1
    else:
        author = blue(username)

    spaces = " " * spacing
    yield f"{author} {spaces} {yellow(time)}"

    yield ""
    yield from html_lines(status.content, width)

    if status.media_attachments:
        yield ""
        yield "Media:"
        for attachment in status.media_attachments:
            url = attachment.url
            for line in wc_wrap(url, width):
                yield line

    if status.poll:
        yield ""
        yield from poll_lines(status.poll, width)

    reblogged_by_acct = f"@{reblogged_by.acct}" if reblogged_by else None
    yield ""

    reply = f"↲ In reply to {yellow(in_reply_to_id)} " if in_reply_to_id else ""
    boost = f"↻ {blue(reblogged_by_acct)} boosted " if reblogged_by else ""
    yield f"ID {yellow(status_id)}  Visibility: {status.visibility}  {reply} {boost}"


def html_lines(html: str, width: int) -> t.Generator[str, None, None]:
    first = True
    for paragraph in html_to_paragraphs(html):
        if not first:
            yield ""
        for line in paragraph:
            for subline in wc_wrap(line, width):
                yield subline
        first = False


def print_poll(poll: Poll):
    for line in poll_lines(poll, get_width()):
        print(line)


def poll_lines(poll: Poll, width: int) -> t.Generator[str, None, None]:
    for idx, option in enumerate(poll.options):
        perc = (
            round(100 * option.votes_count / poll.votes_count)
            if poll.votes_count and option.votes_count is not None
            else 0
        )

        if poll.voted and poll.own_votes and idx in poll.own_votes:
            voted_for = bold("  ✓ Your vote")
        else:
            voted_for = ""

        yield f"{option.title}{voted_for}"

        yield render_poll_bar(f"{perc}% ({option.votes_count})", width, perc)
        yield ""

    poll_footer = f"Poll {poll.id} · {poll.votes_count} votes"

    if poll.expired:
        poll_footer += " · Closed"

    if poll.expires_at:
        expires_at = poll.expires_at.strftime("%Y-%m-%d %H:%M")
        poll_footer += f" · Closes on {expires_at}"

    yield dim(poll_footer)


def render_poll_bar(text: str, width: int, perc: int):
    text = text.ljust(width)
    offset = perc * width // 100
    return click.style(text[:offset], fg="bright_white", bg="blue") + text[offset:].strip()


def print_timeline(items: t.Iterable[Status]):
    print_divider()
    for item in items:
        print_status(item)
        print_divider()


def print_notification(notification: Notification):
    print_notification_header(notification)
    if notification.status:
        print_divider(char="-")
        print_status(notification.status)


def print_notifications(notifications: t.Iterable[Notification]):
    for notification in notifications:
        if notification.type not in ["pleroma:emoji_reaction"]:
            print_divider()
            print_notification(notification)
    print_divider()


def print_notification_header(notification: Notification):
    account_name = format_account_name(notification.account)

    if notification.type == "follow":
        click.echo(f"{account_name} now follows you")
    elif notification.type == "follow_request":
        click.echo(f"{account_name} requested to follow you")
    elif notification.type == "mention":
        click.echo(f"{account_name} mentioned you")
    elif notification.type == "reblog":
        click.echo(f"{account_name} boosted your status")
    elif notification.type == "favourite":
        click.echo(f"{account_name} favourited your status")
    elif notification.type == "update":
        click.echo(f"{account_name} edited a post")
    elif notification.type == "status":
        click.echo(f"{account_name} posted a status")
    elif notification.type == "poll":
        click.echo("A poll you participated in has ended")
    elif notification.type == "admin.sign_up":
        click.echo(f"{account_name} has signed up")
    elif notification.type == "admin.report":
        click.echo(f"{account_name} filed a report")
    else:
        click.secho(
            f"Unknown notification type: '{notification.type}'", err=True, fg="yellow"
        )
        click.secho("Please report an issue to toot.", err=True, fg="yellow")


def print_divider(char: str = "─"):
    click.echo(char * get_width())


def format_tag_name(tag):
    return green(f"#{tag['name']}")


def format_account_name(account: Account) -> str:
    acct = blue(f"@{account.acct}")
    if account.display_name:
        return f"{green(account.display_name)} {acct}"
    else:
        return acct


# Shorthand functions for coloring output


def blue(text: t.Any) -> str:
    return click.style(text, fg="blue")


def bold(text: t.Any) -> str:
    return click.style(text, bold=True)


def cyan(text: t.Any) -> str:
    return click.style(text, fg="cyan")


def dim(text: t.Any) -> str:
    return click.style(text, dim=True)


def green(text: t.Any) -> str:
    return click.style(text, fg="green")


def yellow(text: t.Any) -> str:
    return click.style(text, fg="yellow")



import click
import json as pyjson

from copy import copy
from typing import BinaryIO, Optional

from toot import api
from toot.cli import PRIVACY_CHOICES, cli, json_option, Context, pass_context
from toot.cli.validators import validate_language
from toot.output import print_acct_list


@cli.command()
@click.option("--display-name", help="The display name to use for the profile.")
@click.option("--note", help="The account bio.")
@click.option(
    "--avatar",
    type=click.File(mode="rb"),
    help="Path to the avatar image to set.",
)
@click.option(
    "--header",
    type=click.File(mode="rb"),
    help="Path to the header image to set.",
)
@click.option(
    "--bot/--no-bot",
    default=None,
    help="Whether the account has a bot flag.",
)
@click.option(
    "--discoverable/--no-discoverable",
    default=None,
    help="Whether the account should be shown in the profile directory.",
)
@click.option(
    "--locked/--no-locked",
    default=None,
    help="Whether manual approval of follow requests is required.",
)
@click.option(
    "--privacy",
    type=click.Choice(PRIVACY_CHOICES),
    help="Default post privacy for authored statuses.",
)
@click.option(
    "--sensitive/--no-sensitive",
    default=None,
    help="Whether to mark authored statuses as sensitive by default.",
)
@click.option(
    "--language",
    callback=validate_language,
    help="Default language to use for authored statuses (ISO 639-1).",
)
@json_option
@pass_context
def update_account(
    ctx: Context,
    display_name: Optional[str],
    note: Optional[str],
    avatar: Optional[BinaryIO],
    header: Optional[BinaryIO],
    bot: Optional[bool],
    discoverable: Optional[bool],
    locked: Optional[bool],
    privacy: Optional[bool],
    sensitive: Optional[bool],
    language: Optional[bool],
    json: bool,
):
    """Update your account details"""
    options = [
        avatar,
        bot,
        discoverable,
        display_name,
        header,
        language,
        locked,
        note,
        privacy,
        sensitive,
    ]

    if all(option is None for option in options):
        raise click.ClickException("Please specify at least one option to update the account")

    response = api.update_account(
        ctx.app,
        ctx.user,
        avatar=avatar,
        bot=bot,
        discoverable=discoverable,
        display_name=display_name,
        header=header,
        language=language,
        locked=locked,
        note=note,
        privacy=privacy,
        sensitive=sensitive,
    )

    if json:
        click.echo(response.text)
    else:
        click.secho("✓ Account updated", fg="green")


# Make alias in snake case to keep BC
update_account_alias = copy(update_account)
update_account_alias.hidden = True
cli.add_command(update_account_alias, name="update_account")


@cli.command()
@click.argument("account")
@json_option
@pass_context
def follow(ctx: Context, account: str, json: bool):
    """Follow an account"""
    found_account = api.find_account(ctx.app, ctx.user, account)
    response = api.follow(ctx.app, ctx.user, found_account["id"])
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ You are now following {account}", fg="green")


@cli.command()
@click.argument("account")
@json_option
@pass_context
def unfollow(ctx: Context, account: str, json: bool):
    """Unfollow an account"""
    found_account = api.find_account(ctx.app, ctx.user, account)
    response = api.unfollow(ctx.app, ctx.user, found_account["id"])
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ You are no longer following {account}", fg="green")


@cli.command()
@click.argument("account", required=False)
@json_option
@pass_context
def following(ctx: Context, account: Optional[str], json: bool):
    """List accounts followed by an account.

    If no account is given list accounts followed by you.
    """
    account = account or ctx.user.username
    found_account = api.find_account(ctx.app, ctx.user, account)
    accounts = api.following(ctx.app, ctx.user, found_account["id"])
    if json:
        click.echo(pyjson.dumps(accounts))
    else:
        print_acct_list(accounts)


@cli.command()
@click.argument("account", required=False)
@json_option
@pass_context
def followers(ctx: Context, account: Optional[str], json: bool):
    """List accounts following an account.

    If no account given list accounts following you."""
    account = account or ctx.user.username
    found_account = api.find_account(ctx.app, ctx.user, account)
    accounts = api.followers(ctx.app, ctx.user, found_account["id"])
    if json:
        click.echo(pyjson.dumps(accounts))
    else:
        print_acct_list(accounts)


@cli.command()
@click.argument("account")
@json_option
@pass_context
def mute(ctx: Context, account: str, json: bool):
    """Mute an account"""
    found_account = api.find_account(ctx.app, ctx.user, account)
    response = api.mute(ctx.app, ctx.user, found_account["id"])
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ You have muted {account}", fg="green")


@cli.command()
@click.argument("account")
@json_option
@pass_context
def unmute(ctx: Context, account: str, json: bool):
    """Unmute an account"""
    found_account = api.find_account(ctx.app, ctx.user, account)
    response = api.unmute(ctx.app, ctx.user, found_account["id"])
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ {account} is no longer muted", fg="green")


@cli.command()
@json_option
@pass_context
def muted(ctx: Context, json: bool):
    """List muted accounts"""
    response = api.muted(ctx.app, ctx.user)
    if json:
        click.echo(pyjson.dumps(response))
    else:
        if len(response) > 0:
            click.echo("Muted accounts:")
            print_acct_list(response)
        else:
            click.echo("No accounts muted")


@cli.command()
@click.argument("account")
@json_option
@pass_context
def block(ctx: Context, account: str, json: bool):
    """Block an account"""
    found_account = api.find_account(ctx.app, ctx.user, account)
    response = api.block(ctx.app, ctx.user, found_account["id"])
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ You are now blocking {account}", fg="green")


@cli.command()
@click.argument("account")
@json_option
@pass_context
def unblock(ctx: Context, account: str, json: bool):
    """Unblock an account"""
    found_account = api.find_account(ctx.app, ctx.user, account)
    response = api.unblock(ctx.app, ctx.user, found_account["id"])
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ {account} is no longer blocked", fg="green")


@cli.command()
@json_option
@pass_context
def blocked(ctx: Context, json: bool):
    """List blocked accounts"""
    response = api.blocked(ctx.app, ctx.user)
    if json:
        click.echo(pyjson.dumps(response))
    else:
        if len(response) > 0:
            click.echo("Blocked accounts:")
            print_acct_list(response)
        else:
            click.echo("No accounts blocked")



"""
urWIDgets

A collection of widgets for urwid (https://urwid.org)
"""

__all__ = (
    "parse_text",
    "Hyperlink",
    "TextEmbed",
    # Type Aliases
    "Markup",
    "StringMarkup",
    "ListMarkup",
    "TupleMarkup",
    "NormalTupleMarkup",
    "DisplayAttribute",
    "WidgetTupleMarkup",
    "WidgetListMarkup",
)
__author__ = "Toluwaleke Ogundipe"

from .hyperlink import Hyperlink
from .text_embed import (
    DisplayAttribute,
    ListMarkup,
    Markup,
    NormalTupleMarkup,
    StringMarkup,
    TextEmbed,
    TupleMarkup,
    WidgetListMarkup,
    WidgetTupleMarkup,
    parse_text,
)

version_info = (0, 3, 0, "dev")

# Follows https://semver.org/spec/v2.0.0.html
__version__ = ".".join(map(str, version_info[:3]))
if version_info[3:]:
    __version__ += "-" + ".".join(map(str, version_info[3:]))



"""
Utilities for dealing with string containing wide characters.
"""

import re
from typing import Generator, List

from wcwidth import wcwidth, wcswidth


def _wc_hard_wrap(line: str, length: int) -> Generator[str, None, None]:
    """
    Wrap text to length characters, breaking when target length is reached,
    taking into account character width.

    Used to wrap lines which cannot be wrapped on whitespace.
    """
    chars = []
    chars_len = 0
    for char in line:
        char_len = wcwidth(char)
        if chars_len + char_len > length:
            yield "".join(chars)
            chars: List[str] = []
            chars_len = 0

        chars.append(char)
        chars_len += char_len

    if chars:
        yield "".join(chars)


def wc_wrap(text: str, length: int) -> Generator[str, None, None]:
    """
    Wrap text to given length, breaking on whitespace and taking into account
    character width.

    Meant for use on a single line or paragraph. Will destroy spacing between
    words and paragraphs and any indentation.
    """
    line_words: List[str] = []
    line_len = 0

    words = re.split(r"\s+", text.strip())
    for word in words:
        word_len = wcswidth(word)

        if line_words and line_len + word_len > length:
            line = " ".join(line_words)
            if line_len <= length:
                yield line
            else:
                yield from _wc_hard_wrap(line, length)

            line_words = []
            line_len = 0

        line_words.append(word)
        line_len += word_len + 1  # add 1 to account for space between words

    if line_words:
        line = " ".join(line_words)
        if line_len <= length:
            yield line
        else:
            yield from _wc_hard_wrap(line, length)


def trunc(text: str, length: int) -> str:
    """
    Truncates text to given length, taking into account wide characters.

    If truncated, the last char is replaced by an ellipsis.
    """
    if length < 1:
        raise ValueError("length should be 1 or larger")

    # Remove whitespace first so no unnecessary truncation is done.
    text = text.strip()
    text_length = wcswidth(text)

    if text_length <= length:
        return text

    # We cannot just remove n characters from the end since we don't know how
    # wide these characters are and how it will affect text length.
    # Use wcwidth to determine how many characters need to be truncated.
    chars_to_truncate = 0
    trunc_length = 0
    for char in reversed(text):
        chars_to_truncate += 1
        trunc_length += wcwidth(char)
        if text_length - trunc_length <= length:
            break

    # Additional char to make room for ellipsis
    n = chars_to_truncate + 1
    return text[:-n].strip() + '…'


def pad(text: str, length: int) -> str:
    """Pads text to given length, taking into account wide characters."""
    text_length = wcswidth(text)

    if text_length < length:
        return text + ' ' * (length - text_length)

    return text


def fit_text(text: str, length: int) -> str:
    """Makes text fit the given length by padding or truncating it."""
    text_length = wcswidth(text)

    if text_length > length:
        return trunc(text, length)

    if text_length < length:
        return pad(text, length)

    return text



import re
import urwid
import unicodedata

from bs4.element import NavigableString, Tag
from toot.tui.constants import PALETTE
from toot.urwidgets import Hyperlink, TextEmbed
from toot.utils import parse_html, urlencode_url
from typing import List, Tuple
from urwid.util import decompose_tagmarkup


STYLE_NAMES = [p[0] for p in PALETTE]

# NOTE: update this list if Mastodon starts supporting more block tags
BLOCK_TAGS = ["p", "pre", "li", "blockquote", "h1", "h2", "h3", "h4", "h5", "h6"]


def html_to_widgets(html, recovery_attempt=False) -> List[urwid.Widget]:
    """Convert html to urwid widgets"""
    widgets: List[urwid.Widget] = []
    html = unicodedata.normalize("NFKC", html)
    soup = parse_html(html)

    first_tag = True
    for e in soup.body or soup:
        if isinstance(e, NavigableString):
            if first_tag and not recovery_attempt:
                # if our first "tag" is a navigable string
                # the HTML is out of spec, doesn't start with a tag,
                # we see this in content from Pixelfed servers.
                # attempt a fix by wrapping the HTML with <p></p>
                return html_to_widgets(f"<p>{html}</p>", recovery_attempt=True)
            else:
                continue
        else:
            name = e.name
            # if our HTML starts with a tag, but not a block tag
            # the HTML is out of spec. Attempt a fix by wrapping the
            # HTML with <p></p>
            if (first_tag and not recovery_attempt and name not in BLOCK_TAGS):
                return html_to_widgets(f"<p>{html}</p>", recovery_attempt=True)

            markup = render(name, e)
            first_tag = False

        if not isinstance(markup, urwid.Widget):
            # plaintext, so create a padded text widget
            txt = text_to_widget("", markup)
            markup = urwid.Padding(
                txt,
                align="left",
                width=("relative", 100),
                min_width=None,
            )
        widgets.append(markup)
        # separate top level widgets with a blank line
        widgets.append(urwid.Divider(" "))
    return widgets[:-1]  # but suppress the last blank line


def url_to_widget(url: str):
    try:
        widget = len(url), urwid.Filler(Hyperlink(url, "link", url))
    except ValueError:
        widget = len(url), urwid.Filler(urwid.Text(url))  # don't style as link
    return TextEmbed(widget)


def inline_tag_to_text(tag) -> Tuple:
    """Convert html tag to plain text with tag as attributes recursively"""
    markups = process_inline_tag_children(tag)
    if not markups:
        return (tag.name, "")
    return (tag.name, markups)


def process_inline_tag_children(tag) -> List:
    """Recursively retrieve all children
    and convert to a list of markup text"""
    markups = []
    for child in tag.children:
        if isinstance(child, Tag):
            markup = render(child.name, child)
            markups.append(markup)
        else:
            markups.append(child)
    return markups


URL_PATTERN = re.compile(r"(^.+)\x03(.+$)")


def text_to_widget(attr, markup) -> urwid.Widget:
    markup_list = []
    for run in markup:
        if isinstance(run, tuple):
            txt, attr_list = decompose_tagmarkup(run)
            # find anchor titles with an ETX separator followed by href
            match = URL_PATTERN.match(txt)
            if match:
                label, url = match.groups()
                anchor_attr = get_best_anchor_attr(attr_list)
                try:
                    markup_list.append((
                        len(label),
                        urwid.Filler(Hyperlink(url, anchor_attr, label)),
                    ))
                except ValueError:
                    markup_list.append((
                        len(label),
                        urwid.Filler(urwid.Text(url)),  # don't style as link
                    ))
            else:
                markup_list.append(run)
        else:
            markup_list.append(run)

    return TextEmbed(markup_list)


def process_block_tag_children(tag) -> List[urwid.Widget]:
    """Recursively retrieve all children
    and convert to a list of widgets
    any inline tags containing text will be
    converted to Text widgets"""

    pre_widget_markups = []
    post_widget_markups = []
    child_widgets = []
    found_nested_widget = False

    for child in tag.children:
        if isinstance(child, Tag):
            # child is a nested tag; process using custom method
            # or default to inline_tag_to_text
            result = render(child.name, child)
            if isinstance(result, urwid.Widget):
                found_nested_widget = True
                child_widgets.append(result)
            else:
                if not found_nested_widget:
                    pre_widget_markups.append(result)
                else:
                    post_widget_markups.append(result)
        else:
            # child is text; append to the appropriate markup list
            if not found_nested_widget:
                pre_widget_markups.append(child)
            else:
                post_widget_markups.append(child)

    widget_list = []
    if len(pre_widget_markups):
        widget_list.append(text_to_widget(tag.name, pre_widget_markups))

    if len(child_widgets):
        widget_list += child_widgets

    if len(post_widget_markups):
        widget_list.append(text_to_widget(tag.name, post_widget_markups))

    return widget_list


def get_urwid_attr_name(tag) -> str:
    """Get the class name and translate to a
    name suitable for use as an urwid
    text attribute name"""

    if "class" in tag.attrs:
        clss = tag.attrs["class"]
        if len(clss) > 0:
            style_name = "class_" + "_".join(clss)
            # return the class name, only if we
            # find it as a defined palette name
            if style_name in STYLE_NAMES:
                return style_name

    # fallback to returning the tag name
    return tag.name


def basic_block_tag_handler(tag) -> urwid.Widget:
    """default for block tags that need no special treatment"""
    return urwid.Pile(process_block_tag_children(tag))


def get_best_anchor_attr(attrib_list) -> str:
    if not attrib_list:
        return ""
    flat_al = list(flatten(attrib_list))

    for a in flat_al[0]:
        # ref: https://docs.joinmastodon.org/spec/activitypub/
        # these are the class names (translated to attrib names)
        # that we can support for display

        try:
            if a[0] in ["class_hashtag", "class_mention_hashtag", "class_mention"]:
                return a[0]
        except KeyError:
            continue

    return "a"


def render(attr: str, content: str):
    if attr in ["a"]:
        return render_anchor(content)

    if attr in ["blockquote"]:
        return render_blockquote(content)

    if attr in ["br"]:
        return render_br(content)

    if attr in ["em"]:
        return render_em(content)

    if attr in ["ol"]:
        return render_ol(content)

    if attr in ["pre"]:
        return render_pre(content)

    if attr in ["span"]:
        return render_span(content)

    if attr in ["b", "strong"]:
        return render_strong(content)

    if attr in ["ul"]:
        return render_ul(content)

    # Glitch-soc and Pleroma allow <H1>...<H6> in content
    # Mastodon (PR #23913) does not; header tags are converted to <P><STRONG></STRONG></P>
    if attr in ["p", "div", "li", "h1", "h2", "h3", "h4", "h5", "h6"]:
        return basic_block_tag_handler(content)

    # Fall back to inline_tag_to_text handler
    return inline_tag_to_text(content)


def render_anchor(tag) -> Tuple:
    """anchor tag handler"""

    markups = process_inline_tag_children(tag)
    if not markups:
        return (tag.name, "")

    href = tag.attrs["href"]
    title, attrib_list = decompose_tagmarkup(markups)
    if not attrib_list:
        attrib_list = [tag]
    if href:
        # urlencode the path and query portions of the URL
        href = urlencode_url(href)
        # use ASCII ETX (end of record) as a
        # delimiter between the title and the HREF
        title += f"\x03{href}"

    attr = get_best_anchor_attr(attrib_list)

    if attr == "a":
        # didn't find an attribute to use
        # in the child markup, so let's
        # try the anchor tag's own attributes

        attr = get_urwid_attr_name(tag)

    # hashtag anchors have a class of "mention hashtag"
    # or "hashtag"
    # we'll return style "class_mention_hashtag"
    # or "class_hashtag"
    # in that case; see corresponding palette entry
    # in constants.py controlling hashtag highlighting

    return (attr, title)


def render_blockquote(tag) -> urwid.Widget:
    widget_list = process_block_tag_children(tag)
    blockquote_widget = urwid.LineBox(
        urwid.Padding(
            urwid.Pile(widget_list),
            align="left",
            width=("relative", 100),
            min_width=None,
            left=1,
            right=1,
        ),
        tlcorner="",
        tline="",
        lline="│",
        trcorner="",
        blcorner="",
        rline="",
        bline="",
        brcorner="",
    )
    return urwid.Pile([urwid.AttrMap(blockquote_widget, "blockquote")])


def render_br(tag) -> Tuple:
    return ("br", "\n")


def render_em(tag) -> Tuple:
    # to simplify the number of palette entries
    # translate EM to I (italic)
    markups = process_inline_tag_children(tag)
    if not markups:
        return ("i", "")

    # special case processing for bold and italic
    for parent in tag.parents:
        if parent.name == "b" or parent.name == "strong":
            return ("bi", markups)

    return ("i", markups)


def render_ol(tag) -> urwid.Widget:
    """ordered list tag handler"""

    widgets = []
    list_item_num = 1
    increment = -1 if tag.has_attr("reversed") else 1

    # get ol start= attribute if present
    if tag.has_attr("start") and len(tag.attrs["start"]) > 0:
        try:
            list_item_num = int(tag.attrs["start"])
        except ValueError:
            pass

    for li in tag.find_all("li", recursive=False):
        markup = render("li", li)

        # li value= attribute will change the item number
        # it also overrides any ol start= attribute

        if li.has_attr("value") and len(li.attrs["value"]) > 0:
            try:
                list_item_num = int(li.attrs["value"])
            except ValueError:
                pass

        if not isinstance(markup, urwid.Widget):
            txt = text_to_widget("li", [str(list_item_num), ". ", markup])
            # 1. foo, 2. bar, etc.
            widgets.append(txt)
        else:
            txt = text_to_widget("li", [str(list_item_num), ". "])
            columns = urwid.Columns(
                [txt, ("weight", 9999, markup)], dividechars=1, min_width=3
            )
            widgets.append(columns)

        list_item_num += increment

    return urwid.Pile(widgets)


def render_pre(tag) -> urwid.Widget:
    # <PRE> tag spec says that text should not wrap,
    # but horizontal screen space is at a premium
    # and we have no horizontal scroll bar, so allow
    # wrapping.

    widget_list = [urwid.Divider(" ")]
    widget_list += process_block_tag_children(tag)

    pre_widget = urwid.Padding(
        urwid.Pile(widget_list),
        align="left",
        width=("relative", 100),
        min_width=None,
        left=1,
        right=1,
    )
    return urwid.Pile([urwid.AttrMap(pre_widget, "pre")])


def render_span(tag) -> Tuple:
    markups = process_inline_tag_children(tag)

    if not markups:
        return (tag.name, "")

    # span inherits its parent's class definition
    # unless it has a specific class definition
    # of its own

    if "class" in tag.attrs:
        # uncomment the following code to hide all HTML marked
        # invisible (generally, the http:// prefix of URLs)
        # could be a user preference, it's only advisable if
        # the terminal supports OCS 8 hyperlinks (and that's not
        # automatically detectable)

        # if "invisible" in tag.attrs["class"]:
        #     return (tag.name, "")

        style_name = get_urwid_attr_name(tag)

        if style_name != "span":
            # unique class name matches an entry in our palette
            return (style_name, markups)

    if tag.parent:
        return (get_urwid_attr_name(tag.parent), markups)
    else:
        # fallback
        return ("span", markups)


def render_strong(tag) -> Tuple:
    # to simplify the number of palette entries
    # translate STRONG to B (bold)
    markups = process_inline_tag_children(tag)
    if not markups:
        return ("b", "")

    # special case processing for bold and italic
    for parent in tag.parents:
        if parent.name == "i" or parent.name == "em":
            return ("bi", markups)

    return ("b", markups)


def render_ul(tag) -> urwid.Widget:
    """unordered list tag handler"""

    widgets = []

    for li in tag.find_all("li", recursive=False):
        markup = render("li", li)

        if not isinstance(markup, urwid.Widget):
            txt = text_to_widget("li", ["\N{bullet} ", markup])
            # * foo, * bar, etc.
            widgets.append(txt)
        else:
            txt = text_to_widget("li", ["\N{bullet} "])
            columns = urwid.Columns(
                [txt, ("weight", 9999, markup)], dividechars=1, min_width=3
            )
            widgets.append(columns)

    return urwid.Pile(widgets)


def flatten(data):
    if isinstance(data, tuple):
        for x in data:
            yield from flatten(x)
    else:
        yield data



import json as pyjson
from typing import Optional

import click

from toot import api
from toot.cli import Context, cli, json_option, pass_context
from toot.entities import List, from_dict_list
from toot.output import print_list_accounts, print_lists, print_warning


@cli.group()
def lists():
    """Show and manage your lists"""
    pass


@lists.command()
@json_option
@pass_context
def list(ctx: Context, json: bool):
    """List all your lists"""
    data = api.get_lists(ctx.app, ctx.user)

    if json:
        click.echo(pyjson.dumps(data))
    else:
        if data:
            lists = from_dict_list(List, data)
            print_lists(lists)
        else:
            click.echo("You have no lists defined.")


@lists.command()
@click.argument("title", required=False)
@click.option("--id", help="List ID if not title is given")
@json_option
@pass_context
def accounts(ctx: Context, title: str, id: Optional[str], json: bool):
    """List the accounts in a list"""
    list_id = get_list_id(ctx, title, id)
    response = api.get_list_accounts(ctx.app, ctx.user, list_id)

    if json:
        click.echo(pyjson.dumps(response))
    else:
        print_list_accounts(response)


@lists.command()
@click.argument("title")
@click.option(
    "--replies-policy",
    type=click.Choice(["followed", "list", "none"]),
    default="none",
    help="Replies policy"
)
@json_option
@pass_context
def create(ctx: Context, title: str, replies_policy: str, json: bool):
    """Create a list"""
    response = api.create_list(ctx.app, ctx.user, title=title, replies_policy=replies_policy)
    if json:
        print(response.text)
    else:
        click.secho(f"✓ List \"{title}\" created.", fg="green")


@lists.command()
@click.argument("title", required=False)
@click.option("--id", help="List ID if not title is given")
@json_option
@pass_context
def delete(ctx: Context, title: str, id: Optional[str], json: bool):
    """Delete a list"""
    list_id = get_list_id(ctx, title, id)
    response = api.delete_list(ctx.app, ctx.user, list_id)
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ List \"{title if title else id}\" deleted.", fg="green")


@lists.command()
@click.argument("title", required=False)
@click.argument("account")
@click.option("--id", help="List ID if not title is given")
@json_option
@pass_context
def add(ctx: Context, title: str, account: str, id: Optional[str], json: bool):
    """Add an account to a list"""
    list_id = get_list_id(ctx, title, id)
    found_account = api.find_account(ctx.app, ctx.user, account)

    try:
        response = api.add_accounts_to_list(ctx.app, ctx.user, list_id, [found_account["id"]])
        if json:
            click.echo(response.text)
        else:
            click.secho(f"✓ Added account \"{account}\"", fg="green")
    except Exception:
        # TODO: this is slow, improve
        # if we failed to add the account, try to give a
        # more specific error message than "record not found"
        my_accounts = api.followers(ctx.app, ctx.user, found_account["id"])
        found = False
        if my_accounts:
            for my_account in my_accounts:
                if my_account["id"] == found_account["id"]:
                    found = True
                    break
        if found is False:
            raise click.ClickException(f"You must follow @{account} before adding this account to a list.")
        raise


@lists.command()
@click.argument("title", required=False)
@click.argument("account")
@click.option("--id", help="List ID if not title is given")
@json_option
@pass_context
def remove(ctx: Context, title: str, account: str, id: Optional[str], json: bool):
    """Remove an account from a list"""
    list_id = get_list_id(ctx, title, id)
    found_account = api.find_account(ctx.app, ctx.user, account)
    response = api.remove_accounts_from_list(ctx.app, ctx.user, list_id, [found_account["id"]])
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ Removed account \"{account}\"", fg="green")


# -- Deprecated commands -------------------------------------------------------


@cli.command(name="list_accounts", hidden=True)
@click.argument("title", required=False)
@click.option("--id", help="List ID if not title is given")
@pass_context
def list_accounts(ctx: Context, title: str, id: Optional[str]):
    """List the accounts in a list"""
    print_warning("`toot list_accounts` is deprecated in favour of `toot lists accounts`")
    list_id = get_list_id(ctx, title, id)
    response = api.get_list_accounts(ctx.app, ctx.user, list_id)
    print_list_accounts(response)


@cli.command(name="list_create", hidden=True)
@click.argument("title")
@click.option(
    "--replies-policy",
    type=click.Choice(["followed", "list", "none"]),
    default="none",
    help="Replies policy"
)
@pass_context
def list_create(ctx: Context, title: str, replies_policy: str):
    """Create a list"""
    print_warning("`toot list_create` is deprecated in favour of `toot lists create`")
    api.create_list(ctx.app, ctx.user, title=title, replies_policy=replies_policy)
    click.secho(f"✓ List \"{title}\" created.", fg="green")


@cli.command(name="list_delete", hidden=True)
@click.argument("title", required=False)
@click.option("--id", help="List ID if not title is given")
@pass_context
def list_delete(ctx: Context, title: str, id: Optional[str]):
    """Delete a list"""
    print_warning("`toot list_delete` is deprecated in favour of `toot lists delete`")
    list_id = get_list_id(ctx, title, id)
    api.delete_list(ctx.app, ctx.user, list_id)
    click.secho(f"✓ List \"{title if title else id}\" deleted.", fg="green")


@cli.command(name="list_add", hidden=True)
@click.argument("title", required=False)
@click.argument("account")
@click.option("--id", help="List ID if not title is given")
@pass_context
def list_add(ctx: Context, title: str, account: str, id: Optional[str]):
    """Add an account to a list"""
    print_warning("`toot list_add` is deprecated in favour of `toot lists add`")
    list_id = get_list_id(ctx, title, id)
    found_account = api.find_account(ctx.app, ctx.user, account)

    try:
        api.add_accounts_to_list(ctx.app, ctx.user, list_id, [found_account["id"]])
    except Exception:
        # if we failed to add the account, try to give a
        # more specific error message than "record not found"
        my_accounts = api.followers(ctx.app, ctx.user, found_account["id"])
        found = False
        if my_accounts:
            for my_account in my_accounts:
                if my_account["id"] == found_account["id"]:
                    found = True
                    break
        if found is False:
            raise click.ClickException(f"You must follow @{account} before adding this account to a list.")
        raise

    click.secho(f"✓ Added account \"{account}\"", fg="green")


@cli.command(name="list_remove", hidden=True)
@click.argument("title", required=False)
@click.argument("account")
@click.option("--id", help="List ID if not title is given")
@pass_context
def list_remove(ctx: Context, title: Optional[str], account: str, id: Optional[str]):
    """Remove an account from a list"""
    print_warning("`toot list_remove` is deprecated in favour of `toot lists remove`")
    list_id = get_list_id(ctx, title, id)
    found_account = api.find_account(ctx.app, ctx.user, account)
    api.remove_accounts_from_list(ctx.app, ctx.user, list_id, [found_account["id"]])
    click.secho(f"✓ Removed account \"{account}\"", fg="green")


def get_list_id(ctx: Context, title: Optional[str], list_id: Optional[str]):
    if not list_id and not title:
        raise click.ClickException("Please specify list title or ID")

    lists = api.get_lists(ctx.app, ctx.user)
    matched_ids = [list["id"] for list in lists if _matches(list, title, list_id)]

    if not matched_ids:
        raise click.ClickException("List not found")

    if len(matched_ids) > 1:
        raise click.ClickException("Found multiple lists with the same title, please specify the ID instead")

    return matched_ids[0]


def _matches(list, title, list_id):
    return (title and title.lower() == list["title"].lower()) or (list_id and list_id == list["id"])



import json
import sys

from logging import getLogger
from requests import Request, RequestException, Response
from urllib.parse import urlencode

logger = getLogger("toot")

VERBOSE = "--verbose" in sys.argv


def censor_secrets(headers):
    def _censor(k, v):
        if k == "Authorization":
            return (k, "***CENSORED***")
        return k, v

    return {_censor(k, v) for k, v in headers.items()}


def truncate(line):
    if not VERBOSE and len(line) > 100:
        return line[:100] + "…"

    return line


def log_request(request: Request):
    logger.debug(f" --> {request.method} {_url(request)}")

    if VERBOSE and request.headers:
        headers = censor_secrets(request.headers)
        logger.debug(f" --> HEADERS: {headers}")

    if VERBOSE and request.data:
        data = truncate(request.data)
        logger.debug(f" --> DATA:    {data}")

    if VERBOSE and request.json:
        data = truncate(json.dumps(request.json))
        logger.debug(f" --> JSON:    {data}")

    if VERBOSE and request.files:
        logger.debug(f" --> FILES:   {request.files}")


def log_response(response: Response):
    method = response.request.method
    url = response.request.url
    elapsed = response.elapsed.microseconds // 1000
    logger.debug(f" <-- {method} {url} HTTP {response.status_code} {elapsed}ms")

    if VERBOSE and response.content:
        content = truncate(response.content.decode())
        logger.debug(f" <-- {content}")


def log_request_exception(request: Request, ex: RequestException):
    logger.debug(f" <-- {request.method} {_url(request)} Exception: {ex}")


def _url(request):
    url = request.url
    if request.params:
        url += f"?{urlencode(request.params)}"
    return url



import base64
import re
import sys
import urwid
from collections import OrderedDict
from functools import reduce
from html.parser import HTMLParser
from typing import List

HASHTAG_PATTERN = re.compile(r'(?<!\w)(#\w+)\b')


def highlight_keys(text, high_attr, low_attr=""):
    """
    Takes a string and adds high_attr attribute to parts in square brackets,
    and optionally low_attr attribute to parts outside square brackets.

    The result can be rendered using a urwid.Text widget.

    For example:

    >>> highlight_keys("[P]rint [V]iew", "blue")
    >>> [('blue', 'P'), 'rint ', ('blue', 'V'), 'iew']
    """
    def _gen():
        highlighted = False
        for part in re.split("\\[|\\]", text):
            if part:
                if highlighted:
                    yield (high_attr, part) if high_attr else part
                else:
                    yield (low_attr, part) if low_attr else part
            highlighted = not highlighted
    return list(_gen())


def highlight_hashtags(line):
    hline = []

    for p in re.split(HASHTAG_PATTERN, line):
        if p.startswith("#"):
            hline.append(("hashtag", p))
        else:
            hline.append(p)

    return hline


class LinkParser(HTMLParser):
    def reset(self):
        super().reset()
        self.links = []

    def handle_starttag(self, tag, attrs):
        if tag == "a":
            href, title = None, None
            for name, value in attrs:
                if name == "href":
                    href = value
                if name == "title":
                    title = value
            if href:
                self.links.append((href, title))


def parse_content_links(content):
    """Parse <a> tags from status's `content` and return them as a list of
    (href, title), where `title` may be None.
    """
    parser = LinkParser()
    parser.feed(content)
    return parser.links[:]


def copy_to_clipboard(screen: urwid.display.raw.Screen, text: str):
    """ copy text to clipboard using OSC 52
    This escape sequence is documented
    here https://iterm2.com/documentation-escape-codes.html
    It has wide support - XTerm, Windows Terminal,
    Kitty, iTerm2, others. Some terminals may require a setting to be
    enabled in order to use OSC 52 clipboard functions.
    """

    text_bytes = text.encode("utf-8")
    b64_bytes = base64.b64encode(text_bytes)
    b64_text = b64_bytes.decode("utf-8")

    screen.write(f"\033]52;c;{b64_text}\a")
    screen.flush()


def get_max_toot_chars(instance, default=500):
    # Mastodon
    # https://docs.joinmastodon.org/entities/Instance/#max_characters
    max_toot_chars = deep_get(instance, ["configuration", "statuses", "max_characters"])
    if isinstance(max_toot_chars, int):
        return max_toot_chars

    # Pleroma
    max_toot_chars = instance.get("max_toot_chars")
    if isinstance(max_toot_chars, int):
        return max_toot_chars

    return default


def deep_get(adict: dict, path: List[str], default=None):
    return reduce(
        lambda d, key: d.get(key, default) if isinstance(d, dict) else default,
        path,
        adict
    )


class LRUCache(OrderedDict):
    """Dict with a limited size, ejecting LRUs as needed.
        Default max size = 10Mb"""

    def __init__(self, *args, cache_max_bytes: int = 1024 * 1024 * 10, **kwargs):
        assert cache_max_bytes > 0
        self.total_value_size = 0
        self.cache_max_bytes = cache_max_bytes

        super().__init__(*args, **kwargs)

    def __setitem__(self, key: str, value):
        if key in self:
            self.total_value_size -= sys.getsizeof(super().__getitem__(key).tobytes())
        self.total_value_size += sys.getsizeof(value.tobytes())
        super().__setitem__(key, value)
        super().move_to_end(key)

        while self.total_value_size > self.cache_max_bytes:
            old_key, value = next(iter(self.items()))
            sz = sys.getsizeof(value.tobytes())
            super().__delitem__(old_key)
            self.total_value_size -= sz

    def __getitem__(self, key: str):
        val = super().__getitem__(key)
        super().move_to_end(key)
        return val



import urwid
from wcwidth import wcswidth


class Clickable:
    """
    Add a `click` signal which is sent when the item is activated or clicked.

    TODO: make it work on widgets which have other signals.
    """
    signals = ["click"]

    def keypress(self, size, key):
        if self._command_map[key] == urwid.ACTIVATE:
            self._emit('click')
            return

        return key

    def mouse_event(self, size, event, button, x, y, focus):
        if button == 1:
            self._emit('click')


class SelectableText(Clickable, urwid.Text):
    _selectable = True


class SelectableColumns(Clickable, urwid.Columns):
    _selectable = True


class EditBox(urwid.AttrWrap):
    """Styled edit box."""
    def __init__(self, *args, **kwargs):
        self.edit = urwid.Edit(*args, **kwargs)
        return super().__init__(self.edit, "editbox", "editbox_focused")


class Button(urwid.AttrWrap):
    """Styled button."""
    def __init__(self, *args, **kwargs):
        button = urwid.Button(*args, **kwargs)
        padding = urwid.Padding(button, width=wcswidth(args[0]) + 4)
        return super().__init__(padding, "button", "button_focused")

    def set_label(self, *args, **kwargs):
        self.original_widget.original_widget.set_label(*args, **kwargs)
        self.original_widget.width = wcswidth(args[0]) + 4


class CheckBox(urwid.AttrWrap):
    """Styled checkbox."""
    def __init__(self, *args, **kwargs):
        self.button = urwid.CheckBox(*args, **kwargs)
        padding = urwid.Padding(self.button, width=len(args[0]) + 4)
        return super().__init__(padding, "button", "button_focused")

    def get_state(self):
        """Return the state of the checkbox."""
        return self.button.get_state()


class RadioButton(urwid.AttrWrap):
    """Styled radiobutton."""
    def __init__(self, *args, **kwargs):
        button = urwid.RadioButton(*args, **kwargs)
        padding = urwid.Padding(button, width=len(args[1]) + 4)
        return super().__init__(padding, "button", "button_focused")


class ModalBox(urwid.Frame):
    def __init__(self, message):
        text = urwid.Text(message)
        filler = urwid.Filler(text, valign='top', top=1, bottom=1)
        padding = urwid.Padding(filler, left=1, right=1)
        return super().__init__(padding)


class RoundedLineBox(urwid.LineBox):
    """LineBox that defaults to rounded corners."""
    def __init__(self,
                 original_widget,
                 title="",
                 title_align="center",
                 title_attr=None,
                 tlcorner="\u256d",
                 tline="─",
                 lline="│",
                 trcorner="\u256e",
                 blcorner="\u2570",
                 rline="│",
                 bline="─",
                 brcorner="\u256f",
                 ) -> None:
        return super().__init__(original_widget,
                            title,
                            title_align,
                            title_attr,
                            tlcorner,
                            tline,
                            lline,
                            trcorner,
                            blcorner,
                            rline,
                            bline,
                            brcorner)
